from collections import Counter
import re
from typing import List, Union
from yattag import Doc, indent
from .Header import Header
from .Payload import Payload
from .exceptions import WQXException

class ID(str):
  """The type ID is used for an attribute that uniquely identifies an element in an XML document. An ID value must be an NCName. This means that it must start with a letter or underscore, and can only contain letters, digits, underscores, hyphens, and periods."""

  __pattern = re.compile(r"^[:A-Z_a-z\xC0-\xD6\xD8-\xF6\xF8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][-:0-9.A-Z_a-z\xC0-\xD6\xD8-\xF6\xF8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\xB7\u0300-\u036F\u203F-\u2040]*$")
  def __init__(self, o=None):
    if not isinstance(o, str) or re.match(self.__pattern, o):
      raise ValueError("Attribute of type 'ID' must be a valid XML Identifier.")

class Document:
  """The base document type used for submission to WQXWeb."""

  __id: ID
  __header: Header
  __payload: List[Payload]

  def __init__(self, o=None, *,
    id:ID = None,
    header:Header = None,
    payload:List[Payload] = None
  ):
    if isinstance(o, Document):
      # Assign attributes from object without typechecking
      self.__id = o.id
      self.__header = o.header
      self.__payload = o.payload
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.id = o.get('id', default = None)
      self.header = o.get('header', default = None)
      self.payload = o.get('payload', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.header = header
      self.id = id
      self.payload = payload

  @property
  def header(self) -> Header:
    return self.__header
  @header.setter
  def header(self, val:Header) -> None:
    self.__header = None if val is None else Header(val)

  @property
  def id(self) -> ID:
    return self.__id
  @id.setter
  def id(self, val:ID) -> None:
    self.__id = ID(val)

  @property
  def payload(self) -> List[Payload]:
    return self.__payload
  @payload.setter
  def payload(self, val:Union[Payload,List[Payload]]) -> None:
    if val is None:
      self.__payload = []
    elif isinstance(val, list):
      r:List[Payload] = []
      for x in val:
        r.append(Payload(x))
      self.__payload = r
    else:
      self.__payload = [Payload(val)]

  def list_data_rule_violations(self) -> List[str]:
    """List all data rule (not XSD rules) violations of the enclosed Document.
    This function returns an empty list if none of the tests fail, but that does
    not guarantee the data will be accepted by WQX."""
    violations:List[str] = []

    for payload in self.__payload:
      if payload.operation != payload.UPDATE_INSERT:
        # Data rules only apply for Update-Insert operations
        continue

      data_rule_1 = """
        Data Rule #1: When ElectronicAddressText or ElectronicAddressTypeName is reported, both must be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.electronicAddress is not None )
      ):
        for electronic_address in payload.wqx.organization.electronicAddress:
          if (
            ( electronic_address.electronicAddressText is not None and electronic_address.electronicAddressTypeName is None ) or
            ( electronic_address.electronicAddressText is None and electronic_address.electronicAddressTypeName is not None )
          ):
            violations.append(data_rule_1.strip())

      data_rule_2 = """
        Data Rule #2: When TelephoneNumberText or TelephoneNumberTypeName is reported, both must be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.telephonic is not None )
      ):
        for telephonic in payload.wqx.organization.telephonic:
          if (
            ( telephonic.telephoneNumberText is not None and telephonic.telephoneNumberTypeName is None ) or
            ( telephonic.telephoneNumberText is None and telephonic.telephoneNumberTypeName is not None )
          ):
            violations.append(data_rule_2.strip())

      data_rule_3 = """
        Data Rule #3: When AddressText or AddressTypeName is reported, both must be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.organizationAddress is not None )
      ):
        for organization_address in payload.wqx.organization.organizationAddress:
          if (
            ( organization_address.addressText is not None and organization_address.addressTypeName is None ) or
            ( organization_address.addressText is None and organization_address.addressTypeName is not None )
          ):
            violations.append(data_rule_3.strip())
      
      data_rule_4 = """
        When HorizonitalCollectionMethodName is “Interpolation-Map”, SourceMapScaleNumeric must be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        if payload.wqx.organization.activity is not None:
          for activity in payload.wqx.organization.activity:
            if (
              ( activity.activityLocation is not None ) and
              ( activity.activityLocation.sourceMapScale is None ) and
              ( activity.activityLocation.horizontalCollectionMethodName == 'Interpolation-Map' )
            ):
              violations.append(data_rule_4.strip())
        if payload.wqx.organization.monitoringLocation is not None:
          for monitoring_location in payload.wqx.organization.monitoringLocation:
            if (
              ( monitoring_location.monitoringLocationGeospatial is not None ) and
              ( monitoring_location.monitoringLocationGeospatial.sourceMapScale is None ) and
              ( monitoring_location.monitoringLocationGeospatial.horizontalCollectionMethodName == 'Interpolation-Map' )
            ):
              violations.append(data_rule_4.strip())

      data_rule_5 = """
        When VerticalMeasure's MeasureValue is reported, the following also must be reported:
        VerticalMeasure's MeasureUnitCode,
        VerticalCollectionMethodName,
        VerticalCoordinateReferenceSystemDatumName.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.monitoringLocation is not None )
      ):
        for monitoring_location in payload.wqx.organization.monitoringLocation:
          if (
            ( monitoring_location.monitoringLocationGeospatial is not None ) and
            ( monitoring_location.monitoringLocationGeospatial.verticalMeasure is not None ) and
            ( monitoring_location.monitoringLocationGeospatial.verticalMeasure.measureValue is not None ) and
            (
              ( monitoring_location.monitoringLocationGeospatial.verticalMeasure.measureUnitCode is None ) or
              ( monitoring_location.monitoringLocationGeospatial.verticalCollectionMethodName is None ) or
              ( monitoring_location.monitoringLocationGeospatial.verticalCoordinateReferenceSystemDatumName is None )
            )
          ):
            violations.append(data_rule_5.strip())

      data_rule_6 = """
        Either ProjectDescriptionText or Project's AttachedBinaryObject must be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.project is not None )
      ):
        for project in payload.wqx.organization.project:
          if (
            ( project.projectDescriptionText is None ) and
            ( len (project.attachedBinaryObject) < 1 )
          ):
            violations.append(data_rule_6.strip())

      data_rule_7 = """
        Activity Depth/Height can be reported in only one of the following two ways (but not both):
        a. Specific depth using ActivityDepthHeightMeasure's MeasureValue.
        b. Depth Range using ActivityTopDepthHeightMeasure's MeasureValue and
        ActivityBottomDepthHeightMeasure's MeasureValue.
          i. This method must be used when ActivityTypeCode is “Sample-Integrated Vertical Profile”.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity )
      ):
        for activity in payload.wqx.organization.activity:
          if (
            ( activity.activityDescription is not None ) and
            ( activity.activityDescription.activityDepthHeightMeasure is not None ) and
            ( activity.activityDescription.activityDepthHeightMeasure.measureValue is not None ) and
            (
              ( activity.activityDescription.activityTypeCode == 'Sample-Integrated Vertical Profile' ) or
              (
                ( activity.activityDescription.activityTopDepthHeightMeasure is not None ) and
                ( activity.activityDescription.activityTopDepthHeightMeasure.measureValue is not None )
              ) or (
                ( activity.activityDescription.activityBottomDepthHeightMeasure is not None ) and
                ( activity.activityDescription.activityBottomDepthHeightMeasure.measureValue is not None )
              )
            )
          ):
            violations.append(data_rule_7.strip())
  
      data_rule_8 = """
        When ActivityTypeCode contains the word 'Logger', DataLoggerLineName must be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity )
      ):
        for activity in payload.wqx.organization.activity:
          if (
            ( activity.activityDescription is not None ) and
            ( activity.activityDescription.activityTypeCode is not None ) and
            ( 'Logger' in activity.activityDescription.activityTypeCode )
          ):
            for result in activity.result:
              if result.resultDescription.dataLoggerLineName is None:
                violations.append(data_rule_8.strip())

      data_rule_9 = """
        When ActivityMediaName is "Tissue" then BiologicalIntentName must also be "Tissue" (and visa-versa)
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          for result in activity.result:
            if (
              (
                ( activity.activityDescription is not None ) and
                ( activity.activityDescription.activityMediaName ) and
                ( 'Tissue' != activity.activityDescription.activityMediaName ) and
                ( result.biologicalResultDescription is not None ) and
                ( result.biologicalResultDescription.biologicalIntentName is not None ) and
                ( 'Tissue' == result.biologicalResultDescription.biologicalIntentName )
              ) or
              (
                ( activity.activityDescription is not None ) and
                ( activity.activityDescription.activityMediaName ) and
                ( 'Tissue' == activity.activityDescription.activityMediaName ) and
                ( result.biologicalResultDescription is not None ) and
                ( result.biologicalResultDescription.biologicalIntentName is not None ) and
                ( 'Tissue' != result.biologicalResultDescription.biologicalIntentName )
              )
            ):
              violations.append(data_rule_9.strip())

      data_rule_10 = """
        When ActivityMediaName (or BiologicalIntentName) is "Tissue", then SampleTissueAnatomyName must be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          for result in activity.result:
            if (
              ( activity.activityDescription is not None ) and
              ( activity.activityDescription.activityMediaName is not None ) and
              ( 'Tissue' == activity.activityDescription.activityMediaName ) and
              ( result.biologicalResultDescription is not None ) and
              ( result.biologicalResultDescription.biologicalIntentName is not None ) and
              ( 'Tissue' == result.biologicalResultDescription.biologicalIntentName )
            ):
              if result.biologicalResultDescription.sampleTissueAnatomyName is None:
                violations.append(data_rule_10.strip())

      data_rule_11 = """
        When ActivityMediaName is "Biological" then AssemblageSampledName must be reported
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if (
            ( activity.activityDescription is not None ) and
            ( activity.activityDescription.activityMediaName is not None ) and
            ( 'Biological' == activity.activityDescription.activityMediaName ) and
            (
              ( activity.biologicalActivityDescription is None ) or
              ( activity.biologicalActivityDescription.assemblageSampledName is None )
            )
          ):
            violations.append(data_rule_11.strip())

      data_rule_12 = """
        When ResultDetectionConditionText is 'Not Detected', 'Present Above Quantification Limit' or 'Present Below
        Quantification Limit', then DetectionQuantitationLimitTypeName and DetectionQuantitationLimitMeasure must be
        reported.
        """
      data_rule_12_pattern_1 = ['Not Detected', 'Present Above Quantification Limit', 'Present Below Quantification Limit']
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity )
      ):
        for activity in payload.wqx.organization.activity:
          for result in activity.result:
            if (
              ( result.resultDescription is not None ) and
              ( result.resultDescription.resultDetectionConditionText in data_rule_12_pattern_1 ) and
              ( not any(i.detectionQuantitationLimitTypeName is not None and i.detectionQuantitationLimitMeasure is not None for i in result.resultLabInformation.resultDetectionQuantitationLimit) )
            ):
              violations.append(data_rule_12.strip())

      data_rule_13 = """
        CharacteristicName and ResultStatusIdentifier must be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          for result in activity.result:
            if (
              ( result.resultDescription.characteristicName is None ) or
              ( result.resultDescription.resultStatusIdentifier is None )
            ):
              violations.append(data_rule_13.strip())

      data_rule_14 = """
        When DetectionQuantitationLimit's MeasureValue is reported, DetectionQuantitationLimit's MeasureUnitCode must
        be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.resultLabInformation is not None ) and
                ( result.resultLabInformation.resultDetectionQuantitationLimit is not None )
              ):
                for result_detection_quantitation_limit in result.resultLabInformation.resultDetectionQuantitationLimit:
                  if (
                    ( result_detection_quantitation_limit.detectionQuantitationLimitMeasure is not None ) and
                    ( result_detection_quantitation_limit.detectionQuantitationLimitMeasure.measureValue is not None ) and
                    ( result_detection_quantitation_limit.detectionQuantitationLimitMeasure.measureUnitCode is None )
                  ):
                    violations.append(data_rule_14.strip())

      data_rule_15 = """
        ActivityDescription’s MonitoringLocationIdentifier may be required depending on the value provided for
        ActivityTypeCode. See the domain value list for ActivityTypeCode for more information.
        """
      data_rule_15_domain = []
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if (
            ( activity.activityDescription is not None ) and
            ( activity.activityDescription.monitoringLocationIdentifier is None ) and
            ( activity.activityDescription.activityTypeCode in data_rule_15_domain )
          ):
            violations.append(data_rule_15.strip())

      data_rule_16 = """
        ResultAnalyticalMethod may be required depending on the value provided for ActivityTypeCode. See the domain
        value list for ActivityTypeCode for more information.
        a. However, ResultAnalyticalMethod is never required if BiologicalIntentName is "Individual", "Population
        Census", "Frequency Class", or "Group Summary"
        """
      data_rule_16_pattern = ['Individual', 'Population Census', 'Frequency Class', 'Group Summary']
      data_rule_16_domain = [] # TODO: Add domain values
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.biologicalResultDescription is not None ) and
                ( result.biologicalResultDescription.biologicalIntentName is not None ) and
                ( not result.biologicalResultDescription.biologicalIntentName in data_rule_16_pattern ) and
                ( result.resultAnalyticalMethod is None ) and
                ( activity.activityDescription is not None ) and
                ( activity.activityDescription.activityTypeCode is not None ) and
                ( activity.activityDescription.activityTypeCode in data_rule_16_domain )
              ):
                violations.append(data_rule_16.strip())

      data_rule_17 = """
        ResultSampleFractionText may be required depending on the value provided for CharacteristicName. See the
        domain value list for CharacteristicName for more information.
        """
      data_rule_17_domain = [] # TODO: Add domain values
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.resultDescription is not None ) and
                ( result.resultDescription.resultSampleFractionText is None ) and
                ( result.resultDescription.characteristicName is not None ) and
                ( result.resultDescription.characteristicName in data_rule_17_domain )
              ):
                violations.append(data_rule_17.strip())

      data_rule_18 = """
        ResultAnalyticalMethod’s MethodIdentifierContext must either match a value from the AnalyticalMethodContext
        domain list or it must be the same as the value for the OrganizationIdentifier provided in the submission file.
        a. If the MethodIdentifierContext matches a value from the domain list, then the MethodIdentifier must also
        match a value from the AnalyticalMethod domain list (for that Context). Furthermore, MethodName,
        MethodQualifierTypeName, and MethodDescriptionText are not required and will be ignored (since only the
        Identifier and IdentifierContext are needed to uniquely identify the Analytical Method).
        b. If the MethodIdentifierContext matches your OrganizationIdentifier (indicating your own method), then
        MethodIdentifier and MethodName are both required, but do not need to match a value from the domain list
        (since they are your own). Additionally, MethodQualifierTypeName and MethodDescriptionText can be
        provided, but are optional, to further describe the Analytical Method used."
        """
      data_rule_18_domain = []
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.resultAnalyticalMethod is not None ) and
                ( payload.wqx.organization.organizationDescription is not None ) and
                ( result.resultAnalyticalMethod.methodIdentifierContext != payload.wqx.organization.organizationDescription.organizationIdentifier ) and
                ( not result.resultAnalyticalMethod.methodIdentifierContext in data_rule_18_domain )
              ):
                violations.append(data_rule_18.strip())

      data_rule_19 = """
        ProjectIdentifier, MonitoringLocationIdentifier, ActivityIdentifier, IndexIdentifier and ActivityGroupIdentifier must be
        unique within an Organization. The value for each of these identifiers may occur only once in a submission file.
        a. Unique identifiers are treated as case-insensitive by WQX. For example, the following three identifiers would
        be treated as identical: “Mx571”, “mx571”, “MX571”.
        """
      identifiers:List[str] = []
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None )
      ):
        if payload.wqx.organization.project is not None:
          for project in payload.wqx.organization.project:
            if ( project.projectIdentifier is not None ):
              identifiers.append(project.projectIdentifier.lower())
        if payload.wqx.organization.monitoringLocation is not None:
          for monitoring_location in payload.wqx.organization.monitoringLocation:
            if (
              ( monitoring_location.monitoringLocationIdentity is not None ) and
              ( monitoring_location.monitoringLocationIdentity.monitoringLocationIdentifier is not None )
            ):
              identifiers.append(monitoring_location.monitoringLocationIdentity.monitoringLocationIdentifier.lower())
        if payload.wqx.organization.biologicalHabitatIndex is not None:
          for biological_habitat_index in payload.wqx.organization.biologicalHabitatIndex:
            if biological_habitat_index.indexIdentifier is not None:
              identifiers.append(biological_habitat_index.indexIdentifier.lower())
        if payload.wqx.organization.activity is not None:
          for activity in payload.wqx.organization.activity:
            if (
              ( activity.activityDescription is not None ) and
              ( activity.activityDescription.activityIdentifier is not None )
            ):
              identifiers.append(activity.activityDescription.activityIdentifier.lower())
            if activity.activityMetric is not None:
              for activity_metric in activity.activityMetric:
                if activity_metric.indexIdentifier is not None:
                  identifiers.append(activity_metric.indexIdentifier.lower())
        if payload.wqx.organization.activityGroup is not None:
          for activity_group in payload.wqx.organization.activityGroup:
            if activity_group.activityGroupIdentifier is not None:
              identifiers.append(activity_group.activityGroupIdentifier.lower())
      duplicates:List[str] = [i for i, cnt, in Counter(identifiers).items() if cnt > 1]
      if len(duplicates) > 1:
        violations.append(data_rule_19.strip())

      #pylint: disable=unused-variable
      data_rule_20 = """
        """

      data_rule_21 = """
        ResultMeasure's ResultMeasureValue may be constrained to a list of domain values depending on the value
        provided for CharacteristicName. See the domain value list for CharacteristicName for more information.
        """
      data_rule_21_trigger = []
      data_rule_21_domain = []
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.resultDescription is not None ) and
                ( result.resultDescription.characteristicName in data_rule_21_trigger ) and
                ( result.resultDescription.resultMeasure.resultMeasureValue not in data_rule_21_domain )
              ):
                violations.append(data_rule_21.strip())

      data_rule_22 = """
        If a numeric value is reported for ResultMeasureValue, then ResultMeasure's MeasureUnitCode and
        ResultValueTypeName are required.
        a. The exception to this is when the ResultMeasureValue is a Characteristic Pick List Value. These do not
        have units.
        """
      data_rule_22_exceptions = []
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.resultDescription is not None ) and
                ( result.resultDescription.resultMeasure is not None ) and
                ( result.resultDescription.resultMeasure.resultMeasureValue is not None ) and
                ( result.resultDescription.resultMeasure.resultMeasureValue not in data_rule_22_exceptions ) and
                ( result.resultDescription.resultMeasure.resultMeasureValue.isnumeric() ) and
                (
                  ( result.resultDescription.resultMeasure.measureUnitCode is None ) or
                  ( result.resultDescription.resultValueTypeName is None )
                )
              ):
                violations.append(data_rule_22.strip())

      data_rule_23 = """
        If a CountyCode is reported then a StateCode must also be reported.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.monitoringLocation is not None )
      ):
        if payload.wqx.organization.monitoringLocation is not None:
          for monitoring_location in payload.wqx.organization.monitoringLocation:
            if (
              ( monitoring_location.monitoringLocationGeospatial is not None ) and
              ( monitoring_location.monitoringLocationGeospatial.countyCode is not None ) and
              ( monitoring_location.monitoringLocationGeospatial.stateCode is None )
            ):
              violations.append(data_rule_23.strip())
        if payload.wqx.organization.organizationAddress is not None:
          for organization_address in payload.wqx.organization.organizationAddress:
            if (
              ( organization_address.countyCode is not None ) and
              ( organization_address.stateCode is None )
            ):
              violations.append(data_rule_23.strip())

      data_rule_24 = """
        If NetTypeName = "Net/Horizontal Tow" then BoatSpeedMeasure is required.
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if (
            ( activity.biologicalActivityDescription is not None ) and
            ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation is not None ) and
            ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation is not None ) and
            ( 'Net/Horizontal Tow' == activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netTypeName ) and
            ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.boatSpeedMeasure is None )
          ):
            violations.append(data_rule_24.strip())

      #pylint: disable=unused-variable
      data_rule_25 = """
        If NetTypeName is reported then the SampleCollectionEquipmentName must be one that relates to that type of
        equipment.
        """
      # TODO: Figure out how to report violations of this rule

      data_rule_26 = """
        ActivityMetric's MetricTypeIdentifierContext must either match a value from the MetricTypeContext domain list or it
        must be the same as the value for the OrganizationIdentifier provided in the submission file.
        a. If the MetricTypeIdentifierContext matches a value from the domain list, then the MetricTypeIdentifier must
        also match a value from the MetricType domain list (for that Context). Furthermore, MetricTypeName,
        MetricTypeCitation, MetricTypeScaleText, and FormulaDescriptionText are not required and will be ignored
        (since only the Identifier and IdentifierContext are needed to uniquely identify the MetricType).
        b. If the MetricTypeIdentifierContext matches your OrganizationIdentifier (indicating your own metric), then
        MetricTypeIdentifier and MetricTypeName are both required, but do not need to match a value from the
        domain list (since they are your own). Additionally, MetricTypeCitation, MetricTypeScaleText, and
        FormulaDescriptionText can be provided, but are optional, to further describe the Metric Type used.
        """
      data_rule_26_metric_type_context_domain = []
      data_rule_26_metric_type_domain = []
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.activityMetric is not None:
            for activity_metric in activity.activityMetric:
              if activity_metric.activityMetricType is not None:
                if activity_metric.activityMetricType.metricTypeIdentifierContext == payload.wqx.organization.organizationDescription.organizationIdentifier:
                  if (
                    ( activity_metric.activityMetricType.metricTypeIdentifier is None ) or
                    ( activity_metric.activityMetricType.metricTypeName is None )
                  ):
                    violations.append(data_rule_26.strip())
                elif activity_metric.activityMetricType.metricTypeIdentifierContext in data_rule_26_metric_type_context_domain:
                  if activity_metric.activityMetricType.metricTypeIdentifier not in data_rule_26_metric_type_domain:
                    violations.append(data_rule_26.strip())
                else:
                  violations.append(data_rule_26.strip())

      data_rule_27 = """
        If BiologicalIntentName is "Group Summary" then GroupSummaryCount or GroupSummaryWeight must be reported
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.biologicalResultDescription is not None ) and
                ( 'Group Summary' == result.biologicalResultDescription.biologicalIntentName ) and
                (
                  ( result.biologicalResultDescription.groupSummaryCount is None ) or
                  ( result.biologicalResultDescription.groupSummaryWeightMeasure is None )
                )
              ):
                violations.append(data_rule_27.strip())

      data_rule_28 = """
        If BiologicalIntentName is "Frequency Class" then Result's CharacteristicName must be "Count"
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.biologicalResultDescription is not None ) and
                ( 'Frequency Class' == result.biologicalResultDescription.biologicalIntentName ) and
                ( 'Count' != result.resultDescription.characteristicName )
              ):
                violations.append(data_rule_28.strip())

      data_rule_29 = """
        If BiologicalIntentName is "Population Census" then Result's CharacteristicName must be "Count" or "Total Sample
        Weight"
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.biologicalResultDescription is not None ) and
                ( 'Population Census' == result.biologicalResultDescription.biologicalIntentName ) and
                (
                  ( result.resultDescription is None ) or
                  ( result.resultDescription.characteristicName is None ) or
                  ( result.resultDescription.characteristicName not in ['Count', 'Total Sample Weight'])
                )
              ):
                violations.append(data_rule_29.strip())

      data_rule_30 = """
        FrequencyClassDescriptorUnitCode may be required depending on the value provided for
        FrequencyClassDescriptorCode. See the domain value list for FrequencyClassType for more information.
        """
      data_rule_30_domain = []
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.biologicalResultDescription is not None ) and
                ( result.biologicalResultDescription.frequencyClassInformation is not None )
              ):
                for frequency_class_information in result.biologicalResultDescription.frequencyClassInformation:
                  if (
                    ( frequency_class_information.frequencyClassDescriptorUnitCode is None ) and
                    ( frequency_class_information.frequencyClassDescriptorCode is not None ) and
                    ( frequency_class_information.frequencyClassDescriptorCode in data_rule_30_domain )
                  ):
                    violations.append(data_rule_30.strip())

      data_rule_31 = """
        FrequencyClassInformation's LowerClassBoundValue and UpperClassBoundValue may be required depending on
        the value provided for FrequencyClassDescriptorCode. See the domain value list for FrequencyClassType for more
        information
        """
      data_rule_31_domain = []
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.biologicalResultDescription is not None ) and
                ( result.biologicalResultDescription.frequencyClassInformation is not None )
              ):
                for frequency_class_information in result.biologicalResultDescription.frequencyClassInformation:
                  if (
                    (
                      ( frequency_class_information.lowerClassBoundValue is None ) or
                      ( frequency_class_information.upperClassBoundValue is None )
                    ) and
                    ( frequency_class_information.frequencyClassDescriptorCode is not None ) and
                    ( frequency_class_information.frequencyClassDescriptorCode in data_rule_31_domain )
                  ):
                    violations.append(data_rule_31.strip())

      data_rule_32 = """
        Biological Intent Name and Subject Taxonomic Name must be reported when Activity Media Name is "Biological" or
        "Tissue"
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if (
            ( activity.activityDescription is not None ) and
            ( activity.activityDescription.activityMediaName is not None ) and
            ( activity.activityDescription.activityMediaName in ['Biological', 'Tissue'] )
          ):
            for result in activity.result:
              if (
                ( result.biologicalResultDescription is None ) or
                ( result.biologicalResultDescription.biologicalIntentName is None ) or
                ( result.biologicalResultDescription.subjectTaxonomicName is None )
              ):
                violations.append(data_rule_32.strip())

      data_rule_33 = """
        Either Result Measure Value and/or Result Detection Condition Text must be reported
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.resultDescription is None ) or
                ( result.resultDescription.resultMeasure is None ) or
                (
                  ( result.resultDescription.resultMeasure.resultMeasureValue is None ) and
                  ( result.resultDescription.resultDetectionConditionText is None )
                )
              ):
                violations.append(data_rule_33.strip())

      data_rule_34 = """
        Habitat Selection Method is required when Activity Assemblage is "Benthic Macroinvertebrates"
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if (
            ( activity.biologicalActivityDescription is not None ) and
            ( activity.biologicalActivityDescription.assemblageSampledName is not None ) and
            ( 'Benthic Macroinvertebrates' == activity.biologicalActivityDescription.assemblageSampledName ) and
            (
              ( activity.biologicalActivityDescription is None ) or
              ( activity.biologicalActivityDescription.habitatSelectionMethod is None )
            )
          ):
            violations.append(data_rule_34.strip())

      data_rule_35 = """
        Measure Unit is required when Measure Value is supplied
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None )
      ):
        if payload.wqx.organization.activity is not None:
          for activity in payload.wqx.organization.activity:
            if activity.result is not None:
              for result in activity.result:
                if (
                  (
                    ( result.resultDescription is not None ) and
                    ( result.resultDescription.resultMeasure is not None ) and
                    ( result.resultDescription.resultMeasure.resultMeasureValue is not None ) and
                    ( result.resultDescription.resultMeasure.measureUnitCode is None )
                  ) or (
                    ( result.resultDescription is not None ) and
                    ( result.resultDescription.resultDepthHeightMeasure is not None ) and
                    ( result.resultDescription.resultDepthHeightMeasure.measureValue is not None ) and
                    ( result.resultDescription.resultDepthHeightMeasure.measureUnitCode is None )
                  )
                ):
                  violations.append(data_rule_35.strip())
            if (
              (
                ( activity.activityDescription is not None ) and
                (
                  (
                    ( activity.activityDescription.activityDepthHeightMeasure is not None ) and
                    ( activity.activityDescription.activityDepthHeightMeasure.measureValue is not None ) and
                    ( activity.activityDescription.activityDepthHeightMeasure.measureUnitCode is None )
                  ) or (
                    ( activity.activityDescription.activityTopDepthHeightMeasure is not None ) and
                    ( activity.activityDescription.activityTopDepthHeightMeasure.measureValue is not None ) and
                    ( activity.activityDescription.activityTopDepthHeightMeasure.measureUnitCode is None )
                  ) or (
                    ( activity.activityDescription.activityBottomDepthHeightMeasure is not None ) and
                    ( activity.activityDescription.activityBottomDepthHeightMeasure.measureValue is not None ) and
                    ( activity.activityDescription.activityBottomDepthHeightMeasure.measureUnitCode is None )
                  )
                )
              ) or (
                ( activity.activityLocation is not None ) and
                ( activity.activityLocation.horizontalAccuracyMeasure is not None ) and
                ( activity.activityLocation.horizontalAccuracyMeasure.measureValue is not None ) and
                ( activity.activityLocation.horizontalAccuracyMeasure.measureUnitCode is None )
              ) or (
                ( activity.biologicalActivityDescription is not None ) and
                ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation is not None ) and
                (
                  (
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionDuration is not None ) and
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionDuration.measureValue is not None ) and
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionDuration.measureUnitCode is None )
                  ) or (
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionArea is not None ) and
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionArea.measureValue is not None ) and
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionArea.measureUnitCode is None )
                  ) or (
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachLengthMeasure is not None ) and
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachLengthMeasure.measureValue is not None ) and
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachLengthMeasure.measureUnitCode is None )
                  ) or (
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachWidthMeasure is not None ) and
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachWidthMeasure.measureValue is not None ) and
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachWidthMeasure.measureUnitCode is None )
                  ) or (
                    ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation is not None ) and
                    (
                      (
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netSurfaceAreaMeasure is not None ) and
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netSurfaceAreaMeasure.measureValue is not None ) and
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netSurfaceAreaMeasure.measureUnitCode is None )
                      ) or (
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netMeshSizeMeasure is not None ) and
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netMeshSizeMeasure.measureValue is not None ) and
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netMeshSizeMeasure.measureUnitCode is None )
                      ) or (
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.boatSpeedMeasure is not None ) and
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.boatSpeedMeasure.measureValue is not None ) and
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.boatSpeedMeasure.measureUnitCode is None )
                      ) or (
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.currentSpeedMeasure is not None ) and
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.currentSpeedMeasure.measureValue is not None ) and
                        ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.currentSpeedMeasure.measureUnitCode is None )
                      )
                    )
                  )
                )
              )
            ):
              violations.append(data_rule_35.strip())
            if activity.activityMetric is not None:
              for activity_metric in activity.activityMetric:
                if (
                  ( activity_metric.metricValueMeasure is not None ) and
                  ( activity_metric.metricValueMeasure.measureValue is not None ) and
                  ( activity_metric.metricValueMeasure.measureUnitCode is None )
                ):
                  violations.append(data_rule_35.strip())
            if activity.result is not None:
              for result in activity.result:
                if (
                  ( result.resultLabInformation is not None ) and
                  ( result.resultLabInformation.resultDetectionQuantitationLimit is not None )
                ):
                  for detection_quantitation_limit in result.resultLabInformation.resultDetectionQuantitationLimit:
                    if (
                      ( detection_quantitation_limit.detectionQuantitationLimitMeasure is not None ) and
                      ( detection_quantitation_limit.detectionQuantitationLimitMeasure.measureValue is not None ) and
                      ( detection_quantitation_limit.detectionQuantitationLimitMeasure.measureUnitCode is None )
                    ):
                      violations.append(data_rule_35.strip())
                if (
                  ( result.biologicalResultDescription is not None ) and
                  ( result.biologicalResultDescription.groupSummaryWeightMeasure is not None ) and
                  ( result.biologicalResultDescription.groupSummaryWeightMeasure.measureValue is not None ) and
                  ( result.biologicalResultDescription.groupSummaryWeightMeasure.measureUnitCode is None )
                ):
                  violations.append(data_rule_35.strip())
        if payload.wqx.organization.monitoringLocation is not None:
          for monitoring_location in payload.wqx.organization.monitoringLocation:
            if (
              (
                ( monitoring_location.monitoringLocationGeospatial is not None ) and
                (
                  (
                    ( monitoring_location.monitoringLocationGeospatial.horizontalAccuracyMeasure is not None ) and
                    ( monitoring_location.monitoringLocationGeospatial.horizontalAccuracyMeasure.measureValue is not None ) and
                    ( monitoring_location.monitoringLocationGeospatial.horizontalAccuracyMeasure.measureUnitCode is None )
                  ) or (
                    ( monitoring_location.monitoringLocationGeospatial.verticalAccuracyMeasure is not None ) and
                    ( monitoring_location.monitoringLocationGeospatial.verticalAccuracyMeasure.measureValue is not None ) and
                    ( monitoring_location.monitoringLocationGeospatial.verticalAccuracyMeasure.measureUnitCode is None )
                  ) or (
                    ( monitoring_location.monitoringLocationGeospatial.verticalMeasure is not None ) and
                    ( monitoring_location.monitoringLocationGeospatial.verticalMeasure.measureValue is not None ) and
                    ( monitoring_location.monitoringLocationGeospatial.verticalMeasure.measureUnitCode is None )
                  )
                )
              ) or (
                ( monitoring_location.monitoringLocationIdentity is not None ) and
                (
                  (
                    ( monitoring_location.monitoringLocationIdentity.drainageAreaMeasure is not None ) and
                    ( monitoring_location.monitoringLocationIdentity.drainageAreaMeasure.measureValue is not None ) and
                    ( monitoring_location.monitoringLocationIdentity.drainageAreaMeasure.measureUnitCode is None )
                  ) or (
                    ( monitoring_location.monitoringLocationIdentity.contributingDrainageAreaMeasure is not None ) and
                    ( monitoring_location.monitoringLocationIdentity.contributingDrainageAreaMeasure.measureValue is not None ) and
                    ( monitoring_location.monitoringLocationIdentity.contributingDrainageAreaMeasure.measureUnitCode is None )
                  )
                )
              ) or (
                ( monitoring_location.wellInformation is not None ) and
                (
                  (
                    ( monitoring_location.wellInformation.wellHoleDepthMeasure is not None ) and
                    ( monitoring_location.wellInformation.wellHoleDepthMeasure.measureValue is not None ) and
                    ( monitoring_location.wellInformation.wellHoleDepthMeasure.measureUnitCode is None )
                  ) or (
                    ( monitoring_location.wellInformation.wellDepthMeasure is not None ) and
                    ( monitoring_location.wellInformation.wellDepthMeasure.measureValue is not None ) and
                    ( monitoring_location.wellInformation.wellDepthMeasure.measureUnitCode is None )
                  )
                )
              )
            ):
              violations.append(data_rule_35.strip())
        if payload.wqx.organization.project is not None:
          for project in payload.wqx.organization.project:
            for project_monitoring_location_weighting in project.projectMonitoringLocationWeighting:
              if (
                ( project_monitoring_location_weighting.locationWeightingFactorMeasure is not None ) and
                ( project_monitoring_location_weighting.locationWeightingFactorMeasure.measureValue is not None ) and
                ( project_monitoring_location_weighting.locationWeightingFactorMeasure.measureUnitCode is None )
              ):
                violations.append(data_rule_35.strip())

      data_rule_36 = """
        Measure Value is required when Measurement Unit is supplied
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                (
                  ( result.resultDescription is not None ) and
                  ( result.resultDescription.resultMeasure is not None ) and
                  ( result.resultDescription.resultMeasure.resultMeasureValue is None ) and
                  ( result.resultDescription.resultMeasure.measureUnitCode is not None )
                ) or (
                  ( result.resultDescription is not None ) and
                  ( result.resultDescription.resultDepthHeightMeasure is not None ) and
                  ( result.resultDescription.resultDepthHeightMeasure.measureValue is None ) and
                  ( result.resultDescription.resultDepthHeightMeasure.measureUnitCode is not None )
                )
              ):
                violations.append(data_rule_36.strip())
          if (
            (
              ( activity.activityDescription is not None ) and
              (
                (
                  ( activity.activityDescription.activityDepthHeightMeasure is not None ) and
                  ( activity.activityDescription.activityDepthHeightMeasure.measureValue is None ) and
                  ( activity.activityDescription.activityDepthHeightMeasure.measureUnitCode is not None )
                ) or (
                  ( activity.activityDescription.activityTopDepthHeightMeasure is not None ) and
                  ( activity.activityDescription.activityTopDepthHeightMeasure.measureValue is None ) and
                  ( activity.activityDescription.activityTopDepthHeightMeasure.measureUnitCode is not None )
                ) or (
                  ( activity.activityDescription.activityBottomDepthHeightMeasure is not None ) and
                  ( activity.activityDescription.activityBottomDepthHeightMeasure.measureValue is None ) and
                  ( activity.activityDescription.activityBottomDepthHeightMeasure.measureUnitCode is not None )
                )
              )
            ) or (
              ( activity.activityLocation is not None ) and
              ( activity.activityLocation.horizontalAccuracyMeasure is not None ) and
              ( activity.activityLocation.horizontalAccuracyMeasure.measureValue is None ) and
              ( activity.activityLocation.horizontalAccuracyMeasure.measureUnitCode is not None )
            ) or (
              ( activity.biologicalActivityDescription is not None ) and
              ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation is not None ) and
              (
                (
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionDuration is not None ) and
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionDuration.measureValue is None ) and
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionDuration.measureUnitCode is not None )
                ) or (
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionArea is not None ) and
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionArea.measureValue is None ) and
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.collectionArea.measureUnitCode is not None )
                ) or (
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachLengthMeasure is not None ) and
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachLengthMeasure.measureValue is None ) and
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachLengthMeasure.measureUnitCode is not None )
                ) or (
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachWidthMeasure is not None ) and
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachWidthMeasure.measureValue is None ) and
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.reachWidthMeasure.measureUnitCode is not None )
                ) or (
                  ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation is not None ) and
                  (
                    (
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netSurfaceAreaMeasure is not None ) and
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netSurfaceAreaMeasure.measureValue is None ) and
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netSurfaceAreaMeasure.measureUnitCode is not None )
                    ) or (
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netMeshSizeMeasure is not None ) and
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netMeshSizeMeasure.measureValue is None ) and
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.netMeshSizeMeasure.measureUnitCode is not None )
                    ) or (
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.boatSpeedMeasure is not None ) and
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.boatSpeedMeasure.measureValue is None ) and
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.boatSpeedMeasure.measureUnitCode is not None )
                    ) or (
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.currentSpeedMeasure is not None ) and
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.currentSpeedMeasure.measureValue is None ) and
                      ( activity.biologicalActivityDescription.biologicalHabitatCollectionInformation.netInformation.currentSpeedMeasure.measureUnitCode is not None )
                    )
                  )
                )
              )
            )
          ):
            violations.append(data_rule_36.strip())
          if activity.activityMetric is not None:
            for activity_metric in activity.activityMetric:
              if (
                ( activity_metric.metricValueMeasure is not None ) and
                ( activity_metric.metricValueMeasure.measureValue is None ) and
                ( activity_metric.metricValueMeasure.measureUnitCode is not None )
              ):
                violations.append(data_rule_36.strip())
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.resultLabInformation is not None ) and
                ( result.resultLabInformation.resultDetectionQuantitationLimit is not None )
              ):
                for detection_quantitation_limit in result.resultLabInformation.resultDetectionQuantitationLimit:
                  if (
                    ( detection_quantitation_limit.detectionQuantitationLimitMeasure is not None ) and
                    ( detection_quantitation_limit.detectionQuantitationLimitMeasure.measureValue is None ) and
                    ( detection_quantitation_limit.detectionQuantitationLimitMeasure.measureUnitCode is not None )
                  ):
                    violations.append(data_rule_36.strip())
              if (
                (
                  ( result.biologicalResultDescription is not None ) and
                  ( result.biologicalResultDescription.groupSummaryWeightMeasure is not None ) and
                  ( result.biologicalResultDescription.groupSummaryWeightMeasure.measureValue is None ) and
                  ( result.biologicalResultDescription.groupSummaryWeightMeasure.measureUnitCode is not None )
                )
              ):
                violations.append(data_rule_36.strip())
        if payload.wqx.organization.monitoringLocation is not None:
          for monitoring_location in payload.wqx.organization.monitoringLocation:
            if (
              (
                ( monitoring_location.monitoringLocationGeospatial is not None ) and
                (
                  (
                    ( monitoring_location.monitoringLocationGeospatial.horizontalAccuracyMeasure is not None ) and
                    ( monitoring_location.monitoringLocationGeospatial.horizontalAccuracyMeasure.measureValue is None ) and
                    ( monitoring_location.monitoringLocationGeospatial.horizontalAccuracyMeasure.measureUnitCode is not None )
                  ) or (
                    ( monitoring_location.monitoringLocationGeospatial.verticalAccuracyMeasure is not None ) and
                    ( monitoring_location.monitoringLocationGeospatial.verticalAccuracyMeasure.measureValue is None ) and
                    ( monitoring_location.monitoringLocationGeospatial.verticalAccuracyMeasure.measureUnitCode is not None )
                  ) or (
                    ( monitoring_location.monitoringLocationGeospatial.verticalMeasure is not None ) and
                    ( monitoring_location.monitoringLocationGeospatial.verticalMeasure.measureValue is None ) and
                    ( monitoring_location.monitoringLocationGeospatial.verticalMeasure.measureUnitCode is not None )
                  )
                )
              ) or (
                ( monitoring_location.monitoringLocationIdentity is not None ) and
                (
                  (
                    ( monitoring_location.monitoringLocationIdentity.drainageAreaMeasure is not None ) and
                    ( monitoring_location.monitoringLocationIdentity.drainageAreaMeasure.measureValue is None ) and
                    ( monitoring_location.monitoringLocationIdentity.drainageAreaMeasure.measureUnitCode is not None )
                  ) or (
                    ( monitoring_location.monitoringLocationIdentity.contributingDrainageAreaMeasure is not None ) and
                    ( monitoring_location.monitoringLocationIdentity.contributingDrainageAreaMeasure.measureValue is None ) and
                    ( monitoring_location.monitoringLocationIdentity.contributingDrainageAreaMeasure.measureUnitCode is not None )
                  )
                )
              ) or (
                ( monitoring_location.wellInformation is not None ) and
                (
                  (
                    ( monitoring_location.wellInformation.wellHoleDepthMeasure is not None ) and
                    ( monitoring_location.wellInformation.wellHoleDepthMeasure.measureValue is None ) and
                    ( monitoring_location.wellInformation.wellHoleDepthMeasure.measureUnitCode is not None )
                  ) or (
                    ( monitoring_location.wellInformation.wellDepthMeasure is not None ) and
                    ( monitoring_location.wellInformation.wellDepthMeasure.measureValue is None ) and
                    ( monitoring_location.wellInformation.wellDepthMeasure.measureUnitCode is not None )
                  )
                )
              )
            ):
              violations.append(data_rule_35.strip())
        if payload.wqx.organization.project is not None:
          for project in payload.wqx.organization.project:
            if project.projectMonitoringLocationWeighting is not None:
              for project_monitoring_location_weighting in project.projectMonitoringLocationWeighting:
                if (
                  ( project_monitoring_location_weighting.locationWeightingFactorMeasure is not None ) and
                  ( project_monitoring_location_weighting.locationWeightingFactorMeasure.measureValue is None ) and
                  ( project_monitoring_location_weighting.locationWeightingFactorMeasure.measureUnitCode is not None )
                ):
                  violations.append(data_rule_36.strip())

      data_rule_37 = """
        Target Count is required when the Activity Assemblage is "Benthic Macroinvertebrates"
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if (
            ( activity.biologicalActivityDescription is not None ) and
            ( 'Benthic Macroinvertebrates' == activity.biologicalActivityDescription.assemblageSampledName ) and
            ( activity.result is not None )
          ):
            for result in activity.result:
              if (
                ( result.resultDescription is None ) or
                ( result.resultDescription.targetCount is None )
              ):
                violations.append(data_rule_37.strip())

      data_rule_38 = """
        Percent Sample Processed Numeric is required when the Activity Assemblage is "Benthic Macroinvertebrates"
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if (
            ( activity.biologicalActivityDescription is not None ) and
            ( 'Benthic Macroinvertebrates' == activity.biologicalActivityDescription.assemblageSampledName ) and
            ( activity.result is not None )
          ):
            for result in activity.result:
              if (
                ( result.resultDescription is None ) or
                ( result.resultDescription.proportionSampleProcessedNumeric is None )
              ):
                violations.append(data_rule_38.strip())

      data_rule_39 = """
        Percent Sample Processed Numeric must be a positive number between 0 and 1"
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.resultDescription is not None ) and
                ( result.resultDescription.proportionSampleProcessedNumeric is not None ) and
                (
                  ( result.resultDescription.proportionSampleProcessedNumeric < 0 ) or
                  ( result.resultDescription.proportionSampleProcessedNumeric > 1 )
                )
              ):
                violations.append(data_rule_39.strip())

      #pylint: disable=unused-variable
      data_rule_40 = """
        Removed (v3.0) Sample Collection Method is required when Activity Type Code contains the word "Sample"
        """
      # Not enforced because rule indicates it was removed.

      data_rule_41 = """
        Statistical N-Value Numeric must be a positive whole number
        """
      if (
        ( payload.wqx is not None ) and
        ( payload.wqx.organization is not None ) and
        ( payload.wqx.organization.activity is not None )
      ):
        for activity in payload.wqx.organization.activity:
          if activity.result is not None:
            for result in activity.result:
              if (
                ( result.resultDescription is not None ) and
                ( result.resultDescription.statisticalNValueNumeric is not None ) and
                ( result.resultDescription.statisticalNValueNumeric < 0 )
              ):
                violations.append(data_rule_41.strip())

    return violations

  def list_rule_violations(self) -> List[str]:
    violations:List[str] = []
    
    violations.extend(self.list_data_rule_violations())
    violations.extend(self.list_xsd_rule_violations())

    return violations

  def list_xsd_rule_violations(self) -> List[str]:
    """This function is not yet implemented. XSD rule violations will cause
    generateXML to raise a WQXException, which should be handled when called."""
    violations:List[str] = []
    
    return violations

  def generateXML(self, name:str = 'Document') -> str:
    doc = Doc()
    asis = doc.asis
    tag = doc.tag

    violations = self.list_rule_violations()

    if len(violations) > 0:
      raise WQXException(f"The document contains {len(violations)} data rule violations. Use list_rule_violations function for list.")

    if self.__id is None:
      raise WQXException("Attribute 'id' is required.")
    asis('<?xml version="1.0" encoding="UTF-8"?>')
    with tag(name,
      ('Id', self.__id),
      ('xmlns','http://www.exchangenetwork.net/schema/v1.0/ExchangeNetworkDocument.xsd'),
      ('xmlns:xsi','http://www.w3.org/2001/XMLSchema-instance')
    ):
      if self.__header is None:
        raise WQXException("Attribute 'header' is required.")
      asis(self.__header.generateXML('Header'))
      if len(self.__payload) < 1:
        raise WQXException("Attribute 'payload' must be a list of 1 or more Payload objects.")
      for x in self.__payload:
        asis(x.generateXML('Payload'))

    return indent(doc.getvalue(), indentation=' '*2)



from .BiologicalHabitatCollectionInformation import BiologicalHabitatCollectionInformation
from .SimpleContent import (
  AssemblageSampledName,
  HabitatSelectionMethod,
  ToxicityTestType
)
from yattag import Doc

class BiologicalActivityDescription:
  """Allows for the reporting of biological monitoring activities conducted at a Monitoring Location."""

  __assemblageSampledName: AssemblageSampledName
  __biologicalHabitatCollectionInformation: BiologicalHabitatCollectionInformation
  __toxicityTestType: ToxicityTestType
  __habitatSelectionMethod: HabitatSelectionMethod

  def __init__(self, o=None, *,
    assemblageSampledName:AssemblageSampledName = None,
    biologicalHabitatCollectionInformation:BiologicalHabitatCollectionInformation = None,
    toxicityTestType:ToxicityTestType = None,
    habitatSelectionMethod:HabitatSelectionMethod = None
  ):
    if isinstance(o, BiologicalActivityDescription):
      # Assign attributes from object without typechecking
      self.__assemblageSampledName = o.assemblageSampledName
      self.__biologicalHabitatCollectionInformation = o.biologicalHabitatCollectionInformation
      self.__toxicityTestType = o.toxicityTestType
      self.__habitatSelectionMethod = o.habitatSelectionMethod
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.assemblageSampledName = o.get('assemblageSampledName', default = None)
      self.biologicalHabitatCollectionInformation = o.get('biologicalHabitatCollectionInformation', default = None)
      self.toxicityTestType = o.get('toxicityTestType', default = None)
      self.habitatSelectionMethod = o.get('habitatSelectionMethod', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.assemblageSampledName = assemblageSampledName
      self.biologicalHabitatCollectionInformation = biologicalHabitatCollectionInformation
      self.toxicityTestType = toxicityTestType
      self.habitatSelectionMethod = habitatSelectionMethod

  @property
  def assemblageSampledName(self) -> AssemblageSampledName:
    return self.__assemblageSampledName
  @assemblageSampledName.setter
  def assemblageSampledName(self, val:AssemblageSampledName) -> None:
    self.__assemblageSampledName = None if val is None else AssemblageSampledName(val)

  @property
  def biologicalHabitatCollectionInformation(self) -> BiologicalHabitatCollectionInformation:
    return self.__biologicalHabitatCollectionInformation
  @biologicalHabitatCollectionInformation.setter
  def biologicalHabitatCollectionInformation(self, val:BiologicalHabitatCollectionInformation) -> None:
    self.__biologicalHabitatCollectionInformation = None if val is None else BiologicalHabitatCollectionInformation(val)

  @property
  def toxicityTestType(self) -> ToxicityTestType:
    return self.__toxicityTestType
  @toxicityTestType.setter
  def toxicityTestType(self, val:ToxicityTestType) -> None:
    self.__toxicityTestType = None if val is None else ToxicityTestType(val)

  @property
  def habitatSelectionMethod(self) -> HabitatSelectionMethod:
    return self.__habitatSelectionMethod
  @habitatSelectionMethod.setter
  def habitatSelectionMethod(self, val:HabitatSelectionMethod) -> None:
    self.__habitatSelectionMethod = None if val is None else HabitatSelectionMethod(val)

  def generateXML(self, name:str = 'BiologicalActivityDescription') -> str:
    doc = Doc()
    asis = doc.asis
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__assemblageSampledName is not None:
        line('AssemblageSampledName', self.__assemblageSampledName)
      if self.__biologicalHabitatCollectionInformation is not None:
        asis(self.__biologicalHabitatCollectionInformation.generateXML('BiologicalHabitatCollectionInformation'))
      if self.__toxicityTestType is not None:
        line('ToxicityTestType', self.__toxicityTestType)
      if self.__habitatSelectionMethod is not None:
        line('HabitatSelectionMethod', self.__habitatSelectionMethod)

    return doc.getvalue()



from ..exceptions import WQXException
from .SimpleContent import (
  NewIdentifier,
  OldIdentifier,
  OrganizationIdentifier
)
from typing import List, Union
from yattag import Doc

class IdentifierUpdate:
  """Allows a Project Identifier to be changed."""

  __oldIdentifier: OldIdentifier
  __newIdentifier: NewIdentifier

  def __init__(self, o=None, *,
    oldIdentifier:OldIdentifier = None,
    newIdentifier:NewIdentifier = None
  ):
    if isinstance(o, IdentifierUpdate):
      # Assign attributes from object without typechecking
      self.__oldIdentifier = o.oldIdentifier
      self.__newIdentifier = o.newIdentifier
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.oldIdentifier = o.get('oldIdentifier', default = None)
      self.newIdentifier = o.get('newIdentifier', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.oldIdentifier = oldIdentifier
      self.newIdentifier = newIdentifier

  @property
  def oldIdentifier(self) -> OldIdentifier:
    return self.__oldIdentifier
  @oldIdentifier.setter
  def oldIdentifier(self, val:OldIdentifier) -> None:
    self.__oldIdentifier = None if val is None else OldIdentifier(val)

  @property
  def newIdentifier(self) -> NewIdentifier:
    return self.__newIdentifier
  @newIdentifier.setter
  def newIdentifier(self, val:NewIdentifier) -> None:
    self.__newIdentifier = None if val is None else NewIdentifier(val)

  def generateXML(self, name = 'IdentifierUpdate'):
    doc = Doc()
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__oldIdentifier is None:
        raise WQXException("Attribute 'oldIdentifier' is required.")
      line('OldIdentifier', self.__oldIdentifier)
      if self.__newIdentifier is None:
        raise WQXException("Attribute 'newIdentifier' is required.")
      line('NewIdentifier', self.__newIdentifier)

    return doc.getvalue()

class UpdateIdentifiers:
  """Allows a set of identifiers to be changed."""

  __organizationIdentifier: OrganizationIdentifier
  __projectIdentifierUpdate: List[IdentifierUpdate]
  __monitoringLocationIdentifierUpdate: List[IdentifierUpdate]
  __indexIdentifierUpdate: List[IdentifierUpdate]
  __activityIdentifierUpdate: List[IdentifierUpdate]
  __activityGroupIdentifierUpdate: List[IdentifierUpdate]

  def __init__(self, o=None, *,
    organizationIdentifier:OrganizationIdentifier = None,
    projectIdentifierUpdate:List[IdentifierUpdate] = [],
    monitoringLocationIdentifierUpdate:List[IdentifierUpdate] = [],
    indexIdentifierUpdate:List[IdentifierUpdate] = [],
    activityIdentifierUpdate:List[IdentifierUpdate] = [],
    activityGroupIdentifierUpdate:List[IdentifierUpdate] = []
  ):
    if isinstance(o, UpdateIdentifiers):
      # Assign attributes from object without typechecking
      self.__organizationIdentifier = o.organizationIdentifier
      self.__projectIdentifierUpdate = o.projectIdentifierUpdate
      self.__monitoringLocationIdentifierUpdate = o.monitoringLocationIdentifierUpdate
      self.__indexIdentifierUpdate = o.indexIdentifierUpdate
      self.__activityIdentifierUpdate = o.activityIdentifierUpdate
      self.__activityGroupIdentifierUpdate = o.activityGroupIdentifierUpdate
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.organizationIdentifier = o.get('organizationIdentifier', default = None)
      self.projectIdentifierUpdate = o.get('projectIdentifierUpdate', default = [])
      self.monitoringLocationIdentifierUpdate = o.get('monitoringLocationIdentifierUpdate', default = [])
      self.indexIdentifierUpdate = o.get('indexIdentifierUpdate', default = [])
      self.activityIdentifierUpdate = o.get('activityIdentifierUpdate', default = [])
      self.activityGroupIdentifierUpdate = o.get('activityGroupIdentifierUpdate', default = [])
    else:
      # Assign attributes from named keywords with typechecking
      self.organizationIdentifier = organizationIdentifier
      self.projectIdentifierUpdate = projectIdentifierUpdate
      self.monitoringLocationIdentifierUpdate = monitoringLocationIdentifierUpdate
      self.indexIdentifierUpdate = indexIdentifierUpdate
      self.activityIdentifierUpdate = activityIdentifierUpdate
      self.activityGroupIdentifierUpdate = activityGroupIdentifierUpdate

  @property
  def organizationIdentifier(self) -> OrganizationIdentifier:
    return self.__organizationIdentifier
  @organizationIdentifier.setter
  def organizationIdentifier(self, val:OrganizationIdentifier) -> None:
    self.__organizationIdentifier = OrganizationIdentifier(val)

  @property
  def projectIdentifierUpdate(self) -> List[IdentifierUpdate]:
    return self.__projectIdentifierUpdate
  @projectIdentifierUpdate.setter
  def projectIdentifierUpdate(self, val:Union[IdentifierUpdate,List[IdentifierUpdate]]) -> None:
    if val is None:
      self.__projectIdentifierUpdate = []
    elif isinstance(val, list):
      r:List[IdentifierUpdate] = []
      for x in val:
        r.append(IdentifierUpdate(x))
      self.__projectIdentifierUpdate = r
    else:
      self.__projectIdentifierUpdate = [IdentifierUpdate(val)]

  @property
  def monitoringLocationIdentifierUpdate(self) -> List[IdentifierUpdate]:
    return self.__monitoringLocationIdentifierUpdate
  @monitoringLocationIdentifierUpdate.setter
  def monitoringLocationIdentifierUpdate(self, val:Union[IdentifierUpdate,List[IdentifierUpdate]]) -> None:
    if val is None:
      self.__monitoringLocationIdentifierUpdate = []
    elif isinstance(val, list):
      r:List[IdentifierUpdate] = []
      for x in val:
        r.append(IdentifierUpdate(x))
      self.__monitoringLocationIdentifierUpdate = r
    else:
      self.__monitoringLocationIdentifierUpdate = [IdentifierUpdate(val)]

  @property
  def indexIdentifierUpdate(self) -> List[IdentifierUpdate]:
    return self.__indexIdentifierUpdate
  @indexIdentifierUpdate.setter
  def indexIdentifierUpdate(self, val:Union[IdentifierUpdate,List[IdentifierUpdate]]) -> None:
    if val is None:
      self.__indexIdentifierUpdate = []
    elif isinstance(val, list):
      r:List[IdentifierUpdate] = []
      for x in val:
        r.append(IdentifierUpdate(x))
      self.__indexIdentifierUpdate = r
    else:
      self.__indexIdentifierUpdate = [IdentifierUpdate(val)]

  @property
  def activityIdentifierUpdate(self) -> List[IdentifierUpdate]:
    return self.__activityIdentifierUpdate
  @activityIdentifierUpdate.setter
  def activityIdentifierUpdate(self, val:Union[IdentifierUpdate,List[IdentifierUpdate]]) -> None:
    if val is None:
      self.__activityIdentifierUpdate = []
    elif isinstance(val, list):
      r:List[IdentifierUpdate] = []
      for x in val:
        r.append(IdentifierUpdate(x))
      self.__activityIdentifierUpdate = r
    else:
      self.__activityIdentifierUpdate = [IdentifierUpdate(val)]

  @property
  def activityGroupIdentifierUpdate(self) -> List[IdentifierUpdate]:
    return self.__activityGroupIdentifierUpdate
  @activityGroupIdentifierUpdate.setter
  def activityGroupIdentifierUpdate(self, val:Union[IdentifierUpdate,List[IdentifierUpdate]]) -> None:
    if val is None:
      self.__activityGroupIdentifierUpdate = []
    elif isinstance(val, list):
      r:List[IdentifierUpdate] = []
      for x in val:
        r.append(IdentifierUpdate(x))
      self.__activityGroupIdentifierUpdate = r
    else:
      self.__activityGroupIdentifierUpdate = [IdentifierUpdate(val)]

  def generateXML(self, name = 'UpdateIdentifiers'):
    doc = Doc()
    asis = doc.asis
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__organizationIdentifier is None:
        raise WQXException("Attribute 'organizationIdentifier' is required.")
      line('OrganizationIdentifier', self.__organizationIdentifier)
      for x in self.__projectIdentifierUpdate:
        asis(x.generateXML('ProjectIdentifierUpdate'))
      for x in self.__monitoringLocationIdentifierUpdate:
        asis(x.generateXML('MonitoringLocationIdentifierUpdate'))
      for x in self.__indexIdentifierUpdate:
        asis(x.generateXML('IndexIdentifierUpdate'))
      for x in self.__activityIdentifierUpdate:
        asis(x.generateXML('ActivityIdentifierUpdate'))
      for x in self.__activityGroupIdentifierUpdate:
        asis(x.generateXML('ActivityGroupIdentifierUpdate'))

    return doc.getvalue()



from ..exceptions import WQXException
from .ActivityDescription import ActivityDescription
from .ActivityLocation import ActivityLocation
from .ActivityMetric import ActivityMetric
from .AttachedBinaryObject import AttachedBinaryObject
from .BiologicalActivityDescription import BiologicalActivityDescription
from .Result import Result
from .SampleDescription import SampleDescription
from typing import List, Union
from yattag import Doc

class Activity:
  """Allows for the reporting of monitoring activities conducted at a Monitoring Location."""

  __activityDescription: ActivityDescription
  __activityLocation: ActivityLocation
  __biologicalActivityDescription: BiologicalActivityDescription
  __sampleDescription: SampleDescription
  __activityMetric: List[ActivityMetric]
  __attachedBinaryObject: List[AttachedBinaryObject]
  __result: List[Result]

  def __init__(self, o=None, *,
    activityDescription:ActivityDescription = None,
    activityLocation:ActivityLocation = None,
    biologicalActivityDescription:BiologicalActivityDescription = None,
    sampleDescription:SampleDescription = None,
    activityMetric:List[ActivityMetric] = None,
    attachedBinaryObject:List[AttachedBinaryObject] = None,
    result:List[Result] = None
  ):
    if isinstance(o, Activity):
      # Assign attributes from object without typechecking
      self.__activityDescription = o.activityDescription
      self.__activityLocation = o.activityLocation
      self.__biologicalActivityDescription = o.biologicalActivityDescription
      self.__sampleDescription = o.sampleDescription
      self.__activityMetric = o.activityMetric
      self.__attachedBinaryObject = o.attachedBinaryObject
      self.__result = o.result
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.activityDescription = o.get('activityDescription', default = None)
      self.activityLocation = o.get('activityLocation', default = None)
      self.biologicalActivityDescription = o.get('biologicalActivityDescription', default = None)
      self.sampleDescription = o.get('sampleDescription', default = None)
      self.activityMetric = o.get(activityMetric, default = [])
      self.attachedBinaryObject = o.get('attachedBinaryObject', default = [])
      self.result = o.get('result', default = [])
    else:
      # Assign attributes from named keywords with typechecking
      self.activityDescription = activityDescription
      self.activityLocation = activityLocation
      self.biologicalActivityDescription = biologicalActivityDescription
      self.sampleDescription = sampleDescription
      self.activityMetric = activityMetric
      self.attachedBinaryObject = attachedBinaryObject
      self.result = result

  @property
  def activityDescription(self) -> ActivityDescription:
    return self.__activityDescription
  @activityDescription.setter
  def activityDescription(self, val:ActivityDescription) -> None:
    self.__activityDescription = None if val is None else ActivityDescription(val)

  @property
  def activityLocation(self) -> ActivityLocation:
    return self.__activityLocation
  @activityLocation.setter
  def activityLocation(self, val:ActivityLocation) -> None:
    self.__activityLocation = None if val is None else ActivityLocation(val)

  @property
  def biologicalActivityDescription(self) -> BiologicalActivityDescription:
    return self.__biologicalActivityDescription
  @biologicalActivityDescription.setter
  def biologicalActivityDescription(self, val:BiologicalActivityDescription) -> None:
    self.__biologicalActivityDescription = None if val is None else BiologicalActivityDescription(val)

  @property
  def sampleDescription(self) -> SampleDescription:
    return self.__sampleDescription
  @sampleDescription.setter
  def sampleDescription(self, val:SampleDescription) -> None:
    self.__sampleDescription = None if val is None else SampleDescription(val)

  @property
  def activityMetric(self) -> List[ActivityMetric]:
    return self.__activityMetric
  @activityMetric.setter
  def activityMetric(self, val:Union[ActivityMetric,List[ActivityMetric]]) -> None:
    if val is None:
      self.__activityMetric = []
    elif isinstance(val, list):
      r:List[ActivityMetric] = []
      for x in val:
        r.append(ActivityMetric(x))
      self.__activityMetric = r
    else:
      self.__activityMetric = [ActivityMetric(val)]

  @property
  def attachedBinaryObject(self) -> List[AttachedBinaryObject]:
    return self.__attachedBinaryObject
  @attachedBinaryObject.setter
  def attachedBinaryObject(self, val:Union[AttachedBinaryObject,List[AttachedBinaryObject]]) -> None:
    if val is None:
      self.__attachedBinaryObject = []
    elif isinstance(val, list):
      r:List[AttachedBinaryObject] = []
      for x in val:
        r.append(AttachedBinaryObject(x))
      self.__attachedBinaryObject = r
    else:
      self.__attachedBinaryObject = [AttachedBinaryObject(val)]

  @property
  def result(self) -> List[Result]:
    return self.__result
  @result.setter
  def result(self, val:Union[Result,List[Result]]) -> None:
    if val is None:
      self.__result = []
    elif isinstance(val, list):
      r:List[Result] = []
      for x in val:
        r.append(Result(x))
      self.__result = r
    else:
      self.__result = [Result(val)]

  def generateXML(self, name:str = 'Activity') -> str:
    doc = Doc()
    asis = doc.asis
    tag = doc.tag

    with tag(name):
      if self.__activityDescription is None:
        raise WQXException("Attribute 'activityDescription' is required.")
      asis(self.__activityDescription.generateXML('ActivityDescription'))
      if self.__activityLocation is not None:
        asis(self.__activityLocation.generateXML('ActivityLocation'))
      if self.__biologicalActivityDescription is not None:
        asis(self.__biologicalActivityDescription.generateXML('BiologicalActivityDescription'))
      if self.__sampleDescription is not None:
        asis(self.__sampleDescription.generateXML('SampleDescription'))
      for x in self.__activityMetric:
        asis(x.generateXML('ActivityMetric'))
      for x in self.__attachedBinaryObject:
        asis(x.generateXML('AttachedBinaryObject'))
      for x in self.__result:
        asis(x.generateXML('Result'))

    return doc.getvalue()



from ..exceptions import WQXException
from .AlternateMonitoringLocationIdentity import AlternateMonitoringLocationIdentity
from .MeasureCompact import MeasureCompact
from .SimpleContent import (
  HUCEightDigitCode,
  HUCTwelveDigitCode,
  MonitoringLocationDescriptionText,
  MonitoringLocationIdentifier,
  MonitoringLocationName,
  MonitoringLocationTypeName,
  TribalLandIndicator,
  TribalLandName
)
from typing import List, Union
from yattag import Doc

class MonitoringLocationIdentity:
  """Basic identification information for the location/site that is monitored or used for sampling."""

  __monitoringLocationIdentifier: MonitoringLocationIdentifier
  __monitoringLocationName: MonitoringLocationName
  __monitoringLocationTypeName: MonitoringLocationTypeName
  __monitoringLocationDescriptionText: MonitoringLocationDescriptionText
  __hucEightDigitCode: HUCEightDigitCode
  __hucTwelveDigitCode: HUCTwelveDigitCode
  __tribalLandIndicator: TribalLandIndicator
  __tribalLandName: TribalLandName
  __alternateMonitoringLocationIdentity: List[AlternateMonitoringLocationIdentity]
  __drainageAreaMeasure: MeasureCompact
  __contributingDrainageAreaMeasure: MeasureCompact

  def __init__(self, o=None, *,
    monitoringLocationIdentifier:MonitoringLocationIdentifier = None,
    monitoringLocationName:MonitoringLocationName = None,
    monitoringLocationTypeName:MonitoringLocationTypeName = None,
    monitoringLocationDescriptionText:MonitoringLocationDescriptionText = None,
    hucEightDigitCode:HUCEightDigitCode = None,
    hucTwelveDigitCode:HUCTwelveDigitCode = None,
    tribalLandIndicator:TribalLandIndicator = None,
    tribalLandName:TribalLandName = None,
    alternateMonitoringLocationIdentity:List[AlternateMonitoringLocationIdentity] = None,
    drainageAreaMeasure:MeasureCompact = None,
    contributingDrainageAreaMeasure:MeasureCompact = None
  ):
    if isinstance(o, MonitoringLocationIdentity):
      # Assign attributes from object without typechecking
      self.__monitoringLocationIdentifier = o.monitoringLocationIdentifier
      self.__monitoringLocationName = o.monitoringLocationName
      self.__monitoringLocationTypeName = o.monitoringLocationTypeName
      self.__monitoringLocationDescriptionText = o.monitoringLocationDescriptionText
      self.__hucEightDigitCode = o.hucEightDigitCode
      self.__hucTwelveDigitCode = o.hucTwelveDigitCode
      self.__tribalLandIndicator = o.tribalLandIndicator
      self.__tribalLandName = o.tribalLandName
      self.__alternateMonitoringLocationIdentity = o.alternateMonitoringLocationIdentity
      self.__drainageAreaMeasure = o.drainageAreaMeasure
      self.__contributingDrainageAreaMeasure = o.contributingDrainageAreaMeasure
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.monitoringLocationIdentifier = o.get('monitoringLocationIdentifier', default = None)
      self.monitoringLocationName = o.get('monitoringLocationName', default = None)
      self.monitoringLocationTypeName = o.get('monitoringLocationTypeName', default = None)
      self.monitoringLocationDescriptionText = o.get('monitoringLocationDescriptionText', default = None)
      self.hucEightDigitCode = o.get('hucEightDigitCode', default = None)
      self.hucTwelveDigitCode = o.get('hucTwelveDigitCode', default = None)
      self.tribalLandIndicator = o.get('tribalLandIndicator', default = None)
      self.tribalLandName = o.get('tribalLandName', default = None)
      self.alternateMonitoringLocationIdentity = o.get('alternateMonitoringLocationIdentity', default = None)
      self.drainageAreaMeasure = o.get('drainageAreaMeasure', default = None)
      self.contributingDrainageAreaMeasure = o.get('contributingDrainageAreaMeasure', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.monitoringLocationIdentifier = monitoringLocationIdentifier
      self.monitoringLocationName = monitoringLocationName
      self.monitoringLocationTypeName = monitoringLocationTypeName
      self.monitoringLocationDescriptionText = monitoringLocationDescriptionText
      self.hucEightDigitCode = hucEightDigitCode
      self.hucTwelveDigitCode = hucTwelveDigitCode
      self.tribalLandIndicator = tribalLandIndicator
      self.tribalLandName = tribalLandName
      self.alternateMonitoringLocationIdentity = alternateMonitoringLocationIdentity
      self.drainageAreaMeasure = drainageAreaMeasure
      self.contributingDrainageAreaMeasure = contributingDrainageAreaMeasure

  @property
  def monitoringLocationIdentifier(self) -> MonitoringLocationIdentifier:
    return self.__monitoringLocationIdentifier
  @monitoringLocationIdentifier.setter
  def monitoringLocationIdentifier(self, val:MonitoringLocationIdentifier) -> None:
    self.__monitoringLocationIdentifier = None if val is None else MonitoringLocationIdentifier(val)

  @property
  def monitoringLocationName(self) -> MonitoringLocationName:
    return self.__monitoringLocationName
  @monitoringLocationName.setter
  def monitoringLocationName(self, val:MonitoringLocationName) -> None:
    self.__monitoringLocationName = None if val is None else MonitoringLocationName(val)

  @property
  def monitoringLocationTypeName(self) -> MonitoringLocationTypeName:
    return self.__monitoringLocationTypeName
  @monitoringLocationTypeName.setter
  def monitoringLocationTypeName(self, val:MonitoringLocationTypeName) -> None:
    self.__monitoringLocationTypeName = None if val is None else MonitoringLocationTypeName(val)

  @property
  def monitoringLocationDescriptionText(self) -> MonitoringLocationDescriptionText:
    return self.__monitoringLocationDescriptionText
  @monitoringLocationDescriptionText.setter
  def monitoringLocationDescriptionText(self, val:MonitoringLocationDescriptionText) -> None:
    self.__monitoringLocationDescriptionText = None if val is None else MonitoringLocationDescriptionText(val)

  @property
  def hucEightDigitCode(self) -> HUCEightDigitCode:
    return self.__hucEightDigitCode
  @hucEightDigitCode.setter
  def hucEightDigitCode(self, val:HUCEightDigitCode) -> None:
    self.__hucEightDigitCode = None if val is None else HUCEightDigitCode(val)

  @property
  def hucTwelveDigitCode(self) -> HUCTwelveDigitCode:
    return self.__hucTwelveDigitCode
  @hucTwelveDigitCode.setter
  def hucTwelveDigitCode(self, val:HUCTwelveDigitCode) -> None:
    self.__hucTwelveDigitCode = None if val is None else HUCTwelveDigitCode(val)

  @property
  def tribalLandIndicator(self) -> TribalLandIndicator:
    return self.__tribalLandIndicator
  @tribalLandIndicator.setter
  def tribalLandIndicator(self, val:TribalLandIndicator) -> None:
    self.__tribalLandIndicator = TribalLandIndicator(val)

  @property
  def tribalLandName(self) -> TribalLandName:
    return self.__tribalLandName
  @tribalLandName.setter
  def tribalLandName(self, val:TribalLandName) -> None:
    self.__tribalLandName = None if val is None else TribalLandName(val)

  @property
  def alternateMonitoringLocationIdentity(self) -> List[AlternateMonitoringLocationIdentity]:
    return self.__alternateMonitoringLocationIdentity
  @alternateMonitoringLocationIdentity.setter
  def alternateMonitoringLocationIdentity(self, val:Union[AlternateMonitoringLocationIdentity,List[AlternateMonitoringLocationIdentity]]) -> None:
    if val is None:
      self.__alternateMonitoringLocationIdentity = []
    elif isinstance(val, list):
      r:List[AlternateMonitoringLocationIdentity] = []
      for x in val:
        r.append(AlternateMonitoringLocationIdentity(x))
      self.__alternateMonitoringLocationIdentity = r
    else:
      self.__alternateMonitoringLocationIdentity = [AlternateMonitoringLocationIdentity(val)]

  @property
  def drainageAreaMeasure(self) -> MeasureCompact:
    """The drainage basin of a lake, stream, wetland, or estuary site."""
    return self.__drainageAreaMeasure
  @drainageAreaMeasure.setter
  def drainageAreaMeasure(self, val:MeasureCompact) -> None:
    """The drainage basin of a lake, stream, wetland, or estuary site."""
    self.__drainageAreaMeasure = None if val is None else MeasureCompact(val)

  @property
  def contributingDrainageAreaMeasure(self) -> MeasureCompact:
    """The contributing drainage area of a lake, stream, wetland, or estuary site."""
    return self.__contributingDrainageAreaMeasure
  @contributingDrainageAreaMeasure.setter
  def contributingDrainageAreaMeasure(self, val:MeasureCompact) -> None:
    """The contributing drainage area of a lake, stream, wetland, or estuary site."""
    self.__contributingDrainageAreaMeasure = None if val is None else MeasureCompact(val)

  def generateXML(self, name:str = 'MonitoringLocationIdentity') -> str:
    doc = Doc()
    asis = doc.asis
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__monitoringLocationIdentifier is None:
        raise WQXException("Attribute 'MonitoringLocationIdentifier' is required.")
      line('MonitoringLocationIdentifier', self.__monitoringLocationIdentifier)
      if self.__monitoringLocationName is None:
        raise WQXException("Attribute 'MonitoringLocationName' is required.")
      line('MonitoringLocationName', self.__monitoringLocationName)
      if self.__monitoringLocationTypeName is None:
        raise WQXException("Attribute 'MonitoringLocationTypeName' is required.")
      line('MonitoringLocationTypeName', self.__monitoringLocationTypeName)
      if self.__monitoringLocationDescriptionText is not None:
        line('MonitoringLocationDescriptionText',self.__monitoringLocationDescriptionText)
      if self.__hucEightDigitCode is not None:
        line('HUCEightDigitCode',self.__hucEightDigitCode)
      if self.__hucTwelveDigitCode is not None:
        line('HUCTwelveDigitCode',self.__hucTwelveDigitCode)
      line('TribalLandIndicator',str(self.__tribalLandIndicator))
      if self.__tribalLandName is not None:
        line('TribalLandName',self.__tribalLandName)
      for x in self.__alternateMonitoringLocationIdentity:
        asis(x.generateXML('AlternateMonitoringLocationIdentity'))
      if self.__drainageAreaMeasure is not None:
        asis(self.__drainageAreaMeasure.generateXML('DrainageAreaMeasure'))
      if self.__contributingDrainageAreaMeasure is not None:
        asis(self.__contributingDrainageAreaMeasure.generateXML('ContributingDrainageAreaMeasure'))

    return doc.getvalue()



from ..exceptions import WQXException
from .AttachedBinaryObject import AttachedBinaryObject
from .MonitoringLocationGeospatial import MonitoringLocationGeospatial
from .MonitoringLocationIdentity import MonitoringLocationIdentity
from .WellInformation import WellInformation
from typing import List, Union
from yattag import Doc

class MonitoringLocation:
  """An identifiable location where an environmental sample, onsite measurement, and/or observation is determined."""

  __monitoringLocationIdentity: MonitoringLocationIdentity
  __monitoringLocationGeospatial: MonitoringLocationGeospatial
  __wellInformation: WellInformation
  __attachedBinaryObject: List[AttachedBinaryObject]

  def __init__(self, o=None, *,
    monitoringLocationIdentity:MonitoringLocationIdentity = None,
    monitoringLocationGeospatial:MonitoringLocationGeospatial = None,
    wellInformation:WellInformation = None,
    attachedBinaryObject:List[AttachedBinaryObject] = None
  ):
    if isinstance(o, MonitoringLocation):
      # Assign attributes from object without typechecking
      self.__monitoringLocationIdentity = o.monitoringLocationIdentity
      self.__monitoringLocationGeospatial = o.monitoringLocationGeospatial
      self.__wellInformation = o.wellInformation
      self.__attachedBinaryObject = o.attachedBinaryObject
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.monitoringLocationIdentity = o.get('monitoringLocationIdentity', default = None)
      self.monitoringLocationGeospatial = o.get('monitoringLocationGeospatial', default = None)
      self.wellInformation = o.get('wellInformation', default = None)
      self.attachedBinaryObject = o.get('attachedBinaryObject', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.monitoringLocationIdentity = monitoringLocationIdentity
      self.monitoringLocationGeospatial = monitoringLocationGeospatial
      self.wellInformation = wellInformation
      self.attachedBinaryObject = attachedBinaryObject

  @property
  def monitoringLocationIdentity(self) -> MonitoringLocationIdentity:
    return self.__monitoringLocationIdentity
  @monitoringLocationIdentity.setter
  def monitoringLocationIdentity(self, val:MonitoringLocationIdentity) -> None:
    self.__monitoringLocationIdentity = None if val is None else MonitoringLocationIdentity(val)

  @property
  def monitoringLocationGeospatial(self) -> MonitoringLocationGeospatial:
    return self.__monitoringLocationGeospatial
  @monitoringLocationGeospatial.setter
  def monitoringLocationGeospatial(self, val:MonitoringLocationGeospatial) -> None:
    self.__monitoringLocationGeospatial = None if val is None else MonitoringLocationGeospatial(val)

  @property
  def wellInformation(self) -> WellInformation:
    return self.__wellInformation
  @wellInformation.setter
  def wellInformation(self, val:WellInformation) -> None:
    self.__wellInformation = None if val is None else WellInformation(val)

  @property
  def attachedBinaryObject(self) -> List[AttachedBinaryObject]:
    return self.__attachedBinaryObject
  @attachedBinaryObject.setter
  def attachedBinaryObject(self, val:Union[AttachedBinaryObject,List[AttachedBinaryObject]]) -> None:
    if val is None:
      self.__attachedBinaryObject = []
    elif isinstance(val, list):
      r:List[AttachedBinaryObject] = []
      for x in val:
        r.append(AttachedBinaryObject(x))
      self.__attachedBinaryObject = r
    else:
      self.__attachedBinaryObject = [AttachedBinaryObject(val)]

  def generateXML(self, name:str = 'MonitoringLocation') -> str:
    doc = Doc()
    asis = doc.asis
    tag = doc.tag

    with tag(name):
      if self.__monitoringLocationIdentity is None:
        raise WQXException("Attribute 'MonitoringLocationIdentity' is required.")
      asis(self.__monitoringLocationIdentity.generateXML('MonitoringLocationIdentity'))
      if self.__monitoringLocationGeospatial is None:
        raise WQXException("Attribute 'MonitoringLocationGeospatial' is required.")
      asis(self.__monitoringLocationGeospatial.generateXML('MonitoringLocationGeospatial'))
      if self.__wellInformation is not None:
        asis(self.__wellInformation.generateXML('WellInformation'))
      for x in self.__attachedBinaryObject:
        asis(x.generateXML('AttachedBinaryObject'))

    return doc.getvalue()



from ..exceptions import WQXException
from .SimpleContent import (
  MeasureQualifierCode,
  MeasureUnitCode,
  ResultMeasureValue
)
from typing import List, Union
from yattag import Doc

class Measure:
  """Identifies the value, associated units of measure, and qualifier for measuring the observation or analytical result value."""

  __resultMeasureValue: ResultMeasureValue
  __measureUnitCode: MeasureUnitCode
  __measureQualifierCode: List[MeasureQualifierCode]

  def __init__(self, o=None, *,
    resultMeasureValue:ResultMeasureValue = None,
    measureUnitCode:MeasureUnitCode = None,
    measureQualifierCode:List[MeasureQualifierCode] = []
  ):
    if isinstance(o, Measure):
      # Assign attributes from object without typechecking
      self.__resultMeasureValue = o.resultMeasureValue
      self.__measureUnitCode = o.measureUnitCode
      self.__measureQualifierCode = o.measureQualifierCode
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.__resultMeasureValue = o.get('resultMeasureValue', default = None)
      self.__measureUnitCode = o.get('measureUnitCode', default = None)
      self.__measureQualifierCode = o.get('measureQualifierCode', default = [])
    else:
      # Assign attributes from named keywords with typechecking
      self.__resultMeasureValue = resultMeasureValue
      self.__measureUnitCode = measureUnitCode
      self.__measureQualifierCode = measureQualifierCode

  @property
  def resultMeasureValue(self) -> ResultMeasureValue:
    return self.__resultMeasureValue
  @resultMeasureValue.setter
  def resultMeasureValue(self, val:ResultMeasureValue) -> None:
    self.__resultMeasureValue = None if val is None else ResultMeasureValue(val)

  @property
  def measureUnitCode(self) -> MeasureUnitCode:
    return self.__measureUnitCode
  @measureUnitCode.setter
  def measureUnitCode(self, val:MeasureUnitCode) -> None:
    self.__measureUnitCode = None if val is None else MeasureUnitCode(val)

  @property
  def measureQualifierCode(self) -> List[MeasureQualifierCode]:
    return self.__measureQualifierCode
  @measureQualifierCode.setter
  def measureQualifierCode(self, val:Union[MeasureQualifierCode,List[MeasureQualifierCode]]) -> None:
    if val is None:
      self.__measureQualifierCode = []
    elif isinstance(val, list):
      r:List[MeasureQualifierCode] = []
      for x in val:
        r.append(MeasureQualifierCode(x))
      self.__measureQualifierCode = r
    else:
      self.__measureQualifierCode = [MeasureQualifierCode(val)]

  def generateXML(self, name:str = 'Measure') -> str:
    doc = Doc()
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__resultMeasureValue is not None:
        line('ResultMeasureValue', self.__resultMeasureValue)
      if self.__measureUnitCode is not None:
        line('MeasureUnitCode', self.__measureUnitCode)
      if len(self.__measureQualifierCode) > 6:
        raise WQXException("Attribute 'measureQualifierCode' must be a list of 0 to 6 MeasureQualifierCode objects.")
      for x in self.__measureQualifierCode:
        line('MeasureQualifierCode', x)

    return doc.getvalue()



from ..exceptions import WQXException
from .FrequencyClassInformation import FrequencyClassInformation
from .MeasureCompact import MeasureCompact
from .SimpleContent import (
  BiologicalIndividualIdentifier,
  BiologicalIntentName,
  GroupSummaryCount,
  SampleTissueAnatomyName,
  SubjectTaxonomicName,
  SubjectTaxonomicNameUserSupplied,
  SubjectTaxonomicNameUserSuppliedReferenceText,
  UnidentifiedSpeciesIdentifier
)
from .TaxonomicDetails import TaxonomicDetails
from typing import List, Union
from yattag import Doc

class BiologicalResultDescription:
  """Allows for the reporting of biological result information."""

  __biologicalIntentName: BiologicalIntentName
  __biologicalIndividualIdentifier: BiologicalIndividualIdentifier
  __subjectTaxonomicName: SubjectTaxonomicName
  __subjectTaxonomicNameUserSupplied: SubjectTaxonomicNameUserSupplied
  __subjectTaxonomicNameUserSuppliedReferenceText: SubjectTaxonomicNameUserSuppliedReferenceText
  __unidentifiedSpeciesIdentifier: UnidentifiedSpeciesIdentifier
  __sampleTissueAnatomyName: SampleTissueAnatomyName
  __groupSummaryCount: GroupSummaryCount
  __groupSummaryWeightMeasure: MeasureCompact
  __taxonomicDetails: TaxonomicDetails
  __frequencyClassInformation: List[FrequencyClassInformation]

  def __init__(self, o=None, *,
    biologicalIntentName:BiologicalIntentName = None,
    biologicalIndividualIdentifier:BiologicalIndividualIdentifier = None,
    subjectTaxonomicName:SubjectTaxonomicName = None,
    subjectTaxonomicNameUserSupplied:SubjectTaxonomicNameUserSupplied = None,
    subjectTaxonomicNameUserSuppliedReferenceText:SubjectTaxonomicNameUserSuppliedReferenceText = None,
    unidentifiedSpeciesIdentifier:UnidentifiedSpeciesIdentifier = None,
    sampleTissueAnatomyName:SampleTissueAnatomyName = None,
    groupSummaryCount:GroupSummaryCount = None,
    groupSummaryWeightMeasure:MeasureCompact = None,
    taxonomicDetails:TaxonomicDetails = None,
    frequencyClassInformation:FrequencyClassInformation = None
  ):
    if isinstance(o, BiologicalResultDescription):
      # Assign attributes from object without typechecking
      self.__biologicalIntentName = o.biologicalIntentName
      self.__biologicalIndividualIdentifier = o.biologicalIndividualIdentifier
      self.__subjectTaxonomicName = o.subjectTaxonomicName
      self.__subjectTaxonomicNameUserSupplied = o.subjectTaxonomicNameUserSupplied
      self.__subjectTaxonomicNameUserSuppliedReferenceText = o.subjectTaxonomicNameUserSuppliedReferenceText
      self.__unidentifiedSpeciesIdentifier = o.unidentifiedSpeciesIdentifier
      self.__sampleTissueAnatomyName = o.sampleTissueAnatomyName
      self.__groupSummaryCount = o.groupSummaryCount
      self.__groupSummaryWeightMeasure = o.groupSummaryWeightMeasure
      self.__taxonomicDetails = o.taxonomicDetails
      self.__frequencyClassInformation = o.frequencyClassInformation
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.biologicalIntentName = o.get('biologicalIntentName', default = None)
      self.biologicalIndividualIdentifier = o.get('biologicalIndividualIdentifier', default = None)
      self.subjectTaxonomicName = o.get('subjectTaxonomicName', default = None)
      self.subjectTaxonomicNameUserSupplied = o.get('subjectTaxonomicNameUserSupplied', default = None)
      self.subjectTaxonomicNameUserSuppliedReferenceText = o.get('subjectTaxonomicNameUserSuppliedReferenceText', default = None)
      self.unidentifiedSpeciesIdentifier = o.get('unidentifiedSpeciesIdentifier', default = None)
      self.sampleTissueAnatomyName = o.get('sampleTissueAnatomyName', default = None)
      self.groupSummaryCount = o.get('groupSummaryCount', default = None)
      self.groupSummaryWeightMeasure = o.get('groupSummaryWeightMeasure', default = None)
      self.taxonomicDetails = o.get('taxonomicDetails', default = None)
      self.frequencyClassInformation = o.get('frequencyClassInformation', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.biologicalIntentName = biologicalIntentName
      self.biologicalIndividualIdentifier = biologicalIndividualIdentifier
      self.subjectTaxonomicName = subjectTaxonomicName
      self.subjectTaxonomicNameUserSupplied = subjectTaxonomicNameUserSupplied
      self.subjectTaxonomicNameUserSuppliedReferenceText = subjectTaxonomicNameUserSuppliedReferenceText
      self.unidentifiedSpeciesIdentifier = unidentifiedSpeciesIdentifier
      self.sampleTissueAnatomyName = sampleTissueAnatomyName
      self.groupSummaryCount = groupSummaryCount
      self.groupSummaryWeightMeasure = groupSummaryWeightMeasure
      self.taxonomicDetails = taxonomicDetails
      self.frequencyClassInformation = frequencyClassInformation

  @property
  def biologicalIntentName(self) -> BiologicalIntentName:
    return self.__biologicalIntentName
  @biologicalIntentName.setter
  def biologicalIntentName(self, val:BiologicalIntentName) -> None:
    self.__biologicalIntentName = None if val is None else BiologicalIntentName(val)

  @property
  def biologicalIndividualIdentifier(self) -> BiologicalIndividualIdentifier:
    return self.__biologicalIndividualIdentifier
  @biologicalIndividualIdentifier.setter
  def biologicalIndividualIdentifier(self, val:BiologicalIndividualIdentifier) -> None:
    self.__biologicalIndividualIdentifier = None if val is None else BiologicalIndividualIdentifier(val)

  @property
  def subjectTaxonomicName(self) -> SubjectTaxonomicName:
    return self.__subjectTaxonomicName
  @subjectTaxonomicName.setter
  def subjectTaxonomicName(self, val:SubjectTaxonomicName) -> None:
    self.__subjectTaxonomicName = None if val is None else SubjectTaxonomicName(val)

  @property
  def subjectTaxonomicNameUserSupplied(self) -> SubjectTaxonomicNameUserSupplied:
    return self.__subjectTaxonomicNameUserSupplied
  @subjectTaxonomicNameUserSupplied.setter
  def subjectTaxonomicNameUserSupplied(self, val:SubjectTaxonomicNameUserSupplied) -> None:
    self.__subjectTaxonomicNameUserSupplied = None if val is None else SubjectTaxonomicNameUserSupplied(val)

  @property
  def subjectTaxonomicNameUserSuppliedReferenceText(self) -> SubjectTaxonomicNameUserSuppliedReferenceText:
    return self.__subjectTaxonomicNameUserSuppliedReferenceText
  @subjectTaxonomicNameUserSuppliedReferenceText.setter
  def subjectTaxonomicNameUserSuppliedReferenceText(self, val:SubjectTaxonomicNameUserSuppliedReferenceText) -> None:
    self.__subjectTaxonomicNameUserSuppliedReferenceText = None if val is None else SubjectTaxonomicNameUserSuppliedReferenceText(val)

  @property
  def unidentifiedSpeciesIdentifier(self) -> UnidentifiedSpeciesIdentifier:
    return self.__unidentifiedSpeciesIdentifier
  @unidentifiedSpeciesIdentifier.setter
  def unidentifiedSpeciesIdentifier(self, val:UnidentifiedSpeciesIdentifier) -> None:
    self.__unidentifiedSpeciesIdentifier = None if val is None else UnidentifiedSpeciesIdentifier(val)

  @property
  def sampleTissueAnatomyName(self) -> SampleTissueAnatomyName:
    return self.__sampleTissueAnatomyName
  @sampleTissueAnatomyName.setter
  def sampleTissueAnatomyName(self, val:SampleTissueAnatomyName) -> None:
    self.__sampleTissueAnatomyName = None if val is None else SampleTissueAnatomyName(val)

  @property
  def groupSummaryCount(self) -> GroupSummaryCount:
    """Captures the total count or total sample weight for a Group Summary."""
    return self.__groupSummaryCount
  @groupSummaryCount.setter
  def groupSummaryCount(self, val:GroupSummaryCount) -> None:
    """Captures the total count or total sample weight for a Group Summary."""
    self.__groupSummaryCount = None if val is None else GroupSummaryCount(val)

  @property
  def groupSummaryWeightMeasure(self) -> MeasureCompact:
    return self.__groupSummaryWeightMeasure
  @groupSummaryWeightMeasure.setter
  def groupSummaryWeightMeasure(self, val:MeasureCompact) -> None:
    self.__groupSummaryWeightMeasure = None if val is None else MeasureCompact(val)

  @property
  def taxonomicDetails(self) -> TaxonomicDetails:
    return self.__taxonomicDetails
  @taxonomicDetails.setter
  def taxonomicDetails(self, val:TaxonomicDetails) -> None:
    self.__taxonomicDetails = None if val is None else TaxonomicDetails(val)

  @property
  def frequencyClassInformation(self) -> FrequencyClassInformation:
    return self.__frequencyClassInformation
  @frequencyClassInformation.setter
  def frequencyClassInformation(self, val:Union[FrequencyClassInformation,List[FrequencyClassInformation]]) -> None:
    if val is None:
      self.__frequencyClassInformation = []
    elif isinstance(val, list):
      r:List[FrequencyClassInformation] = []
      for x in val:
        r.append(FrequencyClassInformation(x))
      self.__frequencyClassInformation = r
    else:
      self.__frequencyClassInformation = [FrequencyClassInformation(val)]


  def generateXML(self, name:str = 'BiologicalResultDescription') -> str:
    doc = Doc()
    asis = doc.asis
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__biologicalIntentName is None:
        raise WQXException("Attribute 'biologicalIntentName' is required.")
      line('BiologicalIntentName',self.__biologicalIntentName)
      if self.__biologicalIndividualIdentifier is not None:
        line('BiologicalIndividualIdentifier', self.__biologicalIndividualIdentifier)
      if self.__subjectTaxonomicName is None:
        raise WQXException("Attribute 'subjectTaxonomicName' is required.")
      line('SubjectTaxonomicName',self.__subjectTaxonomicName)
      if self.__subjectTaxonomicNameUserSupplied is not None:
        line('SubjectTaxonomicNameUserSupplied', self.__subjectTaxonomicNameUserSupplied)
      if self.__subjectTaxonomicNameUserSuppliedReferenceText is not None:
        line('SubjectTaxonomicNameUserSuppliedReferenceText', self.__subjectTaxonomicNameUserSuppliedReferenceText)
      if self.__unidentifiedSpeciesIdentifier is not None:
        line('UnidentifiedSpeciesIdentifier', self.__unidentifiedSpeciesIdentifier)
      if self.__sampleTissueAnatomyName is not None:
        line('SampleTissueAnatomyName', self.__sampleTissueAnatomyName)
      if self.__groupSummaryCount is not None:
        line('GroupSummaryCount', self.__groupSummaryCount)
      if self.__groupSummaryWeightMeasure is not None:
        asis(self.__groupSummaryWeightMeasure.generateXML('GroupSummaryWeightMeasure'))
      if self.__taxonomicDetails is not None:
        asis(self.__taxonomicDetails.generateXML('TaxonomicDetails'))
      if len(self.__frequencyClassInformation) > 3:
        raise WQXException("Attribute frequencyClassInformation must be a list of 0 to 3 FrequencyClassInformation objects.")
      for x in self.__frequencyClassInformation:
        asis(x.generateXML('FrequencyClassInformation'))

    return doc.getvalue()



from ..exceptions import WQXException
from .SimpleContent import (
  LocalAquiferCode,
  LocalAquiferCodeContext,
  LocalAquiferDescriptionText,
  LocalAquiferName
)
from yattag import Doc

class AquiferInformation:
  """Identifies the procedures, processes, and references required to determine the methods used to obtain a result."""

  __localAquiferCode: LocalAquiferCode
  __localAquiferCodeContext: LocalAquiferCodeContext
  __localAquiferName: LocalAquiferName
  __localAquiferDescriptionText: LocalAquiferDescriptionText

  def __init__(self, o=None, *,
    localAquiferCode:LocalAquiferCode = None,
    localAquiferCodeContext:LocalAquiferCodeContext = None,
    localAquiferName:LocalAquiferName = None,
    localAquiferDescriptionText:LocalAquiferDescriptionText = None
  ):
    if isinstance(o, AquiferInformation):
      # Assign attributes from object without typechecking
      self.__localAquiferCode = o.localAquiferCode
      self.__localAquiferCodeContext = o.localAquiferCodeContext
      self.__localAquiferName = o.localAquiferName
      self.__localAquiferDescriptionText = o.localAquiferDescriptionText
    elif isinstance(o, dict):
      # Assign attribute from dictionary with typechecking
      self.localAquiferCode = o.get('localAquiferCode', default = None)
      self.localAquiferCodeContext = o.get('localAquiferCodeContext', default = None)
      self.localAquiferName = o.get('localAquiferName', default = None)
      self.localAquiferDescriptionText = o.get('localAquiferDescriptionText', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.localAquiferCode = localAquiferCode
      self.localAquiferCodeContext = localAquiferCodeContext
      self.localAquiferName = localAquiferName
      self.localAquiferDescriptionText = localAquiferDescriptionText

  @property
  def localAquiferCode(self) -> LocalAquiferCode:
    return self.__localAquiferCode
  @localAquiferCode.setter
  def localAquiferCode(self, val:LocalAquiferCode) -> None:
    self.__localAquiferCode = LocalAquiferCode(val)

  @property
  def localAquiferCodeContext(self) -> LocalAquiferCodeContext:
    return self.__localAquiferCodeContext
  @localAquiferCodeContext.setter
  def localAquiferCodeContext(self, val:LocalAquiferCodeContext) -> None:
    self.__localAquiferCodeContext = LocalAquiferCodeContext(val)

  @property
  def localAquiferName(self) -> LocalAquiferName:
    return self.__localAquiferName
  @localAquiferName.setter
  def localAquiferName(self, val:LocalAquiferName) -> None:
    self.__localAquiferName = LocalAquiferName(val)

  @property
  def localAquiferDescriptionText(self) -> LocalAquiferDescriptionText:
    return self.__localAquiferDescriptionText
  @localAquiferDescriptionText.setter
  def localAquiferDescriptionText(self, val:LocalAquiferDescriptionText) -> None:
    self.__localAquiferDescriptionText = None if val is None else LocalAquiferDescriptionText(val)

  def generateXML(self, name:str = 'AquiferInformation') -> str:
    doc = Doc()
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__localAquiferCode is None:
        raise WQXException("Attribute 'localAquiferCode' is required.")
      line('LocalAquiferCode', self.__localAquiferCode)
      if self.__localAquiferCodeContext is None:
        raise WQXException("Attribute 'localAquiferCodeContext' is required.")
      line('LocalAquiferCodeContext', self.__localAquiferCodeContext)
      if self.__localAquiferName is None:
        raise WQXException("Attribute 'localAquiferName' is required.")
      line('LocalAquiferName', self.__localAquiferName)
      if self.__localAquiferDescriptionText is not None:
        line('LocalAquiferDescriptionText', self.__localAquiferDescriptionText)

    return doc.getvalue()



from .SimpleContent import (
  TelephoneExtensionNumberText,
  TelephoneNumberText,
  TelephoneNumberTypeName
)
from yattag import Doc

class Telephonic:
  """An identification of a telephone connection."""

  __telephoneNumberText: TelephoneNumberText
  __telephoneNumberTypeName: TelephoneNumberTypeName
  __telephoneExtensionNumberText: TelephoneExtensionNumberText

  def __init__(self, o=None, *,
    telephoneNumberText:TelephoneNumberText = None,
    telephoneNumberTypeName:TelephoneNumberTypeName = None,
    telephoneExtensionNumberText:TelephoneExtensionNumberText = None
  ):
    if isinstance(o, Telephonic):
      # Assign attributes from object without typechecking
      self.__telephoneNumberText = o.telephoneNumberText
      self.__telephoneNumberTypeName = o.telephoneNumberTypeName
      self.__telephoneExtensionNumberText = o.telephoneExtensionNumberText
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.telephoneNumberText = o.get('telephoneNumberText', default = None)
      self.telephoneNumberTypeName = o.get('telephoneNumberTypeName', default = None)
      self.telephoneExtensionNumberText = o.get('telephoneExtensionNumberText', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.telephoneNumberText = telephoneNumberText
      self.telephoneNumberTypeName = telephoneNumberTypeName
      self.telephoneExtensionNumberText = telephoneExtensionNumberText

  @property
  def telephoneNumberText(self) -> TelephoneNumberText:
    return self.__telephoneNumberText
  @telephoneNumberText.setter
  def telephoneNumberText(self, val:TelephoneNumberText) -> None:
    self.__telephoneNumberText = None if val is None else TelephoneNumberText(val)

  @property
  def telephoneNumberTypeName(self) -> TelephoneNumberTypeName:
    return self.__telephoneNumberTypeName
  @telephoneNumberTypeName.setter
  def telephoneNumberTypeName(self, val:TelephoneNumberTypeName) -> None:
    self.__telephoneNumberTypeName = None if val is None else TelephoneNumberTypeName(val)

  @property
  def telephoneExtensionNumberText(self) -> TelephoneExtensionNumberText:
    return self.__telephoneExtensionNumberText
  @telephoneExtensionNumberText.setter
  def telephoneExtensionNumberText(self, val:TelephoneExtensionNumberText) -> None:
    self.__telephoneExtensionNumberText = None if val is None else TelephoneExtensionNumberText(val)

  def generateXML(self, name:str = 'Telephonic') -> str:
    doc = Doc()
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__telephoneNumberText is not None:
        line('TelephoneNumberText', self.__telephoneNumberText)
      if self.__telephoneNumberTypeName is not None:
        line('TelephoneNumberTypeName', self.__telephoneNumberTypeName)
      if self.__telephoneExtensionNumberText is not None:
        line('TelephoneExtensionNumberText', self.__telephoneExtensionNumberText)

    return doc.getvalue()



from .SimpleContent import (
  AddressText,
  AddressTypeName,
  CountryCode,
  CountyCode,
  LocalityName,
  PostalCode,
  StateCode,
  SupplementalAddressText
)
from yattag import Doc

class OrganizationAddress:
  """The physical address of an organization."""

  __addressTypeName: AddressTypeName
  __addressText: AddressText
  __supplementalAddressText: SupplementalAddressText
  __localityName: LocalityName
  __stateCode: StateCode
  __postalCode: PostalCode
  __countryCode: CountryCode
  __countyCode: CountyCode

  def __init__(self, o=None, *,
    addressTypeName:AddressTypeName = None,
    addressText:AddressText = None,
    supplementalAddressText:SupplementalAddressText = None,
    localityName:LocalityName = None,
    stateCode:StateCode = None,
    postalCode:PostalCode = None,
    countryCode:CountryCode = None,
    countyCode:CountyCode = None
  ):
    if isinstance(o, OrganizationAddress):
      # Assign attributes from object without typechecking
      self.__addressTypeName = o.addressTypeName
      self.__addressText = o.addressText
      self.__supplementalAddressText = o.supplementalAddressText
      self.__localityName = o.localityName
      self.__stateCode = o.stateCode
      self.__postalCode = o.postalCode
      self.__countryCode = o.countryCode
      self.__countyCode = o.countyCode
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.addressTypeName = o.get('addressTypeName', default = None)
      self.addressText = o.get('addressText', default = None)
      self.supplementalAddressText = o.get('supplementalAddressText', default = None)
      self.localityName = o.get('localityName', default = None)
      self.stateCode = o.get('stateCode', default = None)
      self.postalCode = o.get('postalCode', default = None)
      self.countryCode = o.get('countryCode', default = None)
      self.countyCode = o.get('countyCode', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.addressTypeName = addressTypeName
      self.addressText = addressText
      self.supplementalAddressText = supplementalAddressText
      self.localityName = localityName
      self.stateCode = stateCode
      self.postalCode = postalCode
      self.countryCode = countryCode
      self.countyCode = countyCode

  @property
  def addressTypeName(self) -> AddressTypeName:
    return self.__addressTypeName
  @addressTypeName.setter
  def addressTypeName(self, val:AddressTypeName) -> None:
    self.__addressTypeName = None if val is None else AddressTypeName(val)

  @property
  def addressText(self) -> AddressText:
    return self.__addressText
  @addressText.setter
  def addressText(self, val:AddressText) -> None:
    self.__addressText = None if val is None else AddressText(val)

  @property
  def supplementalAddressText(self) -> SupplementalAddressText:
    return self.__supplementalAddressText
  @supplementalAddressText.setter
  def supplementalAddressText(self, val:SupplementalAddressText) -> None:
    self.__supplementalAddressText = None if val is None else SupplementalAddressText(val)

  @property
  def localityName(self) -> LocalityName:
    return self.__localityName
  @localityName.setter
  def localityName(self, val:LocalityName) -> None:
    self.__localityName = None if val is None else LocalityName(val)

  @property
  def stateCode(self) -> StateCode:
    return self.__stateCode
  @stateCode.setter
  def stateCode(self, val:StateCode) -> None:
    self.__stateCode = None if val is None else StateCode(val)

  @property
  def postalCode(self) -> PostalCode:
    return self.__postalCode
  @postalCode.setter
  def postalCode(self, val:PostalCode) -> None:
    self.__postalCode = None if val is None else PostalCode(val)

  @property
  def countryCode(self) -> CountryCode:
    return self.__countryCode
  @countryCode.setter
  def countryCode(self, val:CountryCode) -> None:
    self.__countryCode = None if val is None else CountryCode(val)

  @property
  def countyCode(self) -> CountyCode:
    return self.__countyCode
  @countyCode.setter
  def countyCode(self, val:CountyCode) -> None:
    self.__countyCode = None if val is None else CountyCode(val)

  def generateXML(self, name:str = 'OrganizationAddress') -> str:
    doc = Doc()
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__addressTypeName is not None:
        line('AddressTypeName', self.__addressTypeName)
      if self.__addressText is not None:
        line('AddressText', self.__addressText)
      if self.__supplementalAddressText is not None:
        line('SupplementalAddressText', self.__supplementalAddressText)
      if self.__localityName is not None:
        line('LocalityName', self.__localityName)
      if self.__stateCode is not None:
        line('StateCode', self.__stateCode)
      if self.__postalCode is not None:
        line('PostalCode', self.__postalCode)
      if self.__countryCode is not None:
        line('CountryCode', self.__countryCode)
      if self.__countyCode is not None:
        line('CountyCode', self.__countyCode)

    return doc.getvalue()



from .DataQualityIndicator import DataQuality
from .Measure import Measure
from .MeasureCompact import MeasureCompact
from .SimpleContent import (
  CharacteristicName,
  CharacteristicNameUserSupplied,
  CommentText,
  DataLoggerLineName,
  DepthAltitudeReferencePointText,
  MethodSpeciationName,
  ProportionSampleProcessedNumeric,
  RecordIdentifierUserSupplied,
  ResultDetectionConditionText,
  ResultParticleSizeBasisText,
  ResultSampleFractionText,
  ResultSamplingPointCommentText,
  ResultSamplingPointName,
  ResultSamplingPointPlaceInSeries,
  ResultSamplingPointType,
  ResultStatusIdentifier,
  ResultTemperatureBasisText,
  ResultTimeBasisText,
  ResultValueTypeName,
  ResultWeightBasisText,
  StatisticalBaseCode,
  StatisticalNValueNumeric,
  TargetCount
)
from yattag import Doc

class ResultDescription:
  """Describes the results of a field measurement, observation, or laboratory analysis."""

  __dataLoggerLineName: DataLoggerLineName
  __resultDetectionConditionText: ResultDetectionConditionText
  __characteristicName: CharacteristicName
  __characteristicNameUserSupplied: CharacteristicNameUserSupplied
  __methodSpeciationName: MethodSpeciationName
  __resultSampleFractionText: ResultSampleFractionText
  __resultMeasure: Measure
  __targetCount: TargetCount
  __proportionSampleProcessedNumeric: ProportionSampleProcessedNumeric
  __resultStatusIdentifier: ResultStatusIdentifier
  __statisticalBaseCode: StatisticalBaseCode
  __statisticalNValueNumeric: StatisticalNValueNumeric
  __resultValueTypeName: ResultValueTypeName
  __resultWeightBasisText: ResultWeightBasisText
  __resultTimeBasisText: ResultTimeBasisText
  __resultTemperatureBasisText: ResultTemperatureBasisText
  __resultParticleSizeBasisText: ResultParticleSizeBasisText
  __dataQuality: DataQuality
  __resultCommentText: CommentText
  __resultDepthHeightMeasure: MeasureCompact
  __resultDepthAltitudeReferencePointText: DepthAltitudeReferencePointText
  __resultSamplingPointName: ResultSamplingPointName
  __resultSamplingPointType: ResultSamplingPointType
  __resultSamplingPointPlaceInSeries: ResultSamplingPointPlaceInSeries
  __resultSamplingPointCommentText: ResultSamplingPointCommentText
  __recordIdentifierUserSupplied: RecordIdentifierUserSupplied

  def __init__(self, o=None, *,
    dataLoggerLineName:DataLoggerLineName = None,
    resultDetectionConditionText:ResultDetectionConditionText = None,
    characteristicName:CharacteristicName = None,
    characteristicNameUserSupplied:CharacteristicNameUserSupplied = None,
    methodSpeciationName:MethodSpeciationName = None,
    resultSampleFractionText:ResultSampleFractionText = None,
    resultMeasure:Measure = None,
    targetCount:TargetCount = None,
    proportionSampleProcessedNumeric:ProportionSampleProcessedNumeric = None,
    resultStatusIdentifier:ResultStatusIdentifier = None,
    statisticalBaseCode:StatisticalBaseCode = None,
    statisticalNValueNumeric:StatisticalNValueNumeric = None,
    resultValueTypeName:ResultValueTypeName = None,
    resultWeightBasisText:ResultWeightBasisText = None,
    resultTimeBasisText:ResultTimeBasisText = None,
    resultTemperatureBasisText:ResultTemperatureBasisText = None,
    resultParticleSizeBasisText:ResultParticleSizeBasisText = None,
    dataQuality:DataQuality = None,
    resultCommentText:CommentText = None,
    resultDepthHeightMeasure:MeasureCompact = None,
    resultDepthAltitudeReferencePointText:DepthAltitudeReferencePointText = None,
    resultSamplingPointName:ResultSamplingPointName = None,
    resultSamplingPointType:ResultSamplingPointType = None,
    resultSamplingPointPlaceInSeries:ResultSamplingPointPlaceInSeries = None,
    resultSamplingPointCommentText:ResultSamplingPointCommentText = None,
    recordIdentifierUserSupplied:RecordIdentifierUserSupplied = None
  ):
    if isinstance(o, ResultDescription):
      # Assign attributes from object without typechecking
      self.__dataLoggerLineName = o.dataLoggerLineName
      self.__resultDetectionConditionText = o.resultDetectionConditionText
      self.__characteristicName = o.characteristicName
      self.__characteristicNameUserSupplied = o.characteristicNameUserSupplied
      self.__methodSpeciationName = o.methodSpeciationName
      self.__resultSampleFractionText = o.resultSampleFractionText
      self.__resultMeasure = o.resultMeasure
      self.__targetCount = o.targetCount
      self.__proportionSampleProcessedNumeric = o.proportionSampleProcessedNumeric
      self.__resultStatusIdentifier = o.resultStatusIdentifier
      self.__statisticalBaseCode = o.statisticalBaseCode
      self.__statisticalNValueNumeric = o.statisticalNValueNumeric
      self.__resultValueTypeName = o.resultValueTypeName
      self.__resultWeightBasisText = o.resultWeightBasisText
      self.__resultTimeBasisText = o.resultTimeBasisText
      self.__resultTemperatureBasisText = o.resultTemperatureBasisText
      self.__resultParticleSizeBasisText = o.resultParticleSizeBasisText
      self.__dataQuality = o.dataQuality
      self.__resultCommentText = o.resultCommentText
      self.__resultDepthHeightMeasure = o.resultDepthHeightMeasure
      self.__resultDepthAltitudeReferencePointText = o.resultDepthAltitudeReferencePointText
      self.__resultSamplingPointName = o.resultSamplingPointName
      self.__resultSamplingPointType = o.resultSamplingPointType
      self.__resultSamplingPointPlaceInSeries = o.resultSamplingPointPlaceInSeries
      self.__resultSamplingPointCommentText = o.resultSamplingPointCommentText
      self.__recordIdentifierUserSupplied = o.recordIdentifierUserSupplied
    elif isinstance(o, dict):
      # Assign attributes from dictionary with typechecking
      self.dataLoggerLineName = o.get('dataLoggerLineName', default = None)
      self.resultDetectionConditionText = o.get('resultDetectionConditionText', default = None)
      self.characteristicName = o.get('characteristicName', default = None)
      self.characteristicNameUserSupplied = o.get('characteristicNameUserSupplied', default = None)
      self.methodSpeciationName = o.get('methodSpeciationName', default = None)
      self.resultSampleFractionText = o.get('resultSampleFractionText', default = None)
      self.resultMeasure = o.get('resultMeasure', default = None)
      self.targetCount = o.get('targetCount', default = None)
      self.proportionSampleProcessedNumeric = o.get('proportionSampleProcessedNumeric', default = None)
      self.resultStatusIdentifier = o.get('resultStatusIdentifier', default = None)
      self.statisticalBaseCode = o.get('statisticalBaseCode', default = None)
      self.statisticalNValueNumeric = o.get('statisticalNValueNumeric', default = None)
      self.resultValueTypeName = o.get('resultValueTypeName', default = None)
      self.resultWeightBasisText = o.get('resultWeightBasisText', default = None)
      self.resultTimeBasisText = o.get('resultTimeBasisText', default = None)
      self.resultTemperatureBasisText = o.get('resultTemperatureBasisText', default = None)
      self.resultParticleSizeBasisText = o.get('resultParticleSizeBasisText', default = None)
      self.dataQuality = o.get('dataQuality', default = None)
      self.resultCommentText = o.get('resultCommentText', default = None)
      self.resultDepthHeightMeasure = o.get('resultDepthHeightMeasure', default = None)
      self.resultDepthAltitudeReferencePointText = o.get('resultDepthAltitudeReferencePointText', default = None)
      self.resultSamplingPointName = o.get('resultSamplingPointName', default = None)
      self.resultSamplingPointType = o.get('resultSamplingPointType', default = None)
      self.resultSamplingPointPlaceInSeries = o.get('resultSamplingPointPlaceInSeries', default = None)
      self.resultSamplingPointCommentText = o.get('resultSamplingPointCommentText', default = None)
      self.recordIdentifierUserSupplied = o.get('recordIdentifierUserSupplied', default = None)
    else:
      # Assign attributes from named keywords with typechecking
      self.dataLoggerLineName = dataLoggerLineName
      self.resultDetectionConditionText = resultDetectionConditionText
      self.characteristicName = characteristicName
      self.characteristicNameUserSupplied = characteristicNameUserSupplied
      self.methodSpeciationName = methodSpeciationName
      self.resultSampleFractionText = resultSampleFractionText
      self.resultMeasure = resultMeasure
      self.targetCount = targetCount
      self.proportionSampleProcessedNumeric = proportionSampleProcessedNumeric
      self.resultStatusIdentifier = resultStatusIdentifier
      self.statisticalBaseCode = statisticalBaseCode
      self.statisticalNValueNumeric = statisticalNValueNumeric
      self.resultValueTypeName = resultValueTypeName
      self.resultWeightBasisText = resultWeightBasisText
      self.resultTimeBasisText = resultTimeBasisText
      self.resultTemperatureBasisText = resultTemperatureBasisText
      self.resultParticleSizeBasisText = resultParticleSizeBasisText
      self.dataQuality = dataQuality
      self.resultCommentText = resultCommentText
      self.resultDepthHeightMeasure = resultDepthHeightMeasure
      self.resultDepthAltitudeReferencePointText = resultDepthAltitudeReferencePointText
      self.resultSamplingPointName = resultSamplingPointName
      self.resultSamplingPointType = resultSamplingPointType
      self.resultSamplingPointPlaceInSeries = resultSamplingPointPlaceInSeries
      self.resultSamplingPointCommentText = resultSamplingPointCommentText
      self.recordIdentifierUserSupplied = recordIdentifierUserSupplied

  @property
  def dataLoggerLineName(self) -> DataLoggerLineName:
    return self.__dataLoggerLineName
  @dataLoggerLineName.setter
  def dataLoggerLineName(self, val:DataLoggerLineName) -> None:
    self.__dataLoggerLineName = None if val is None else DataLoggerLineName(val)

  @property
  def resultDetectionConditionText(self) -> ResultDetectionConditionText:
    return self.__resultDetectionConditionText
  @resultDetectionConditionText.setter
  def resultDetectionConditionText(self, val:ResultDetectionConditionText) -> None:
    self.__resultDetectionConditionText = None if val is None else ResultDetectionConditionText(val)

  @property
  def characteristicName(self) -> CharacteristicName:
    return self.__characteristicName
  @characteristicName.setter
  def characteristicName(self, val:CharacteristicName) -> None:
    self.__characteristicName = None if val is None else CharacteristicName(val)

  @property
  def characteristicNameUserSupplied(self) -> CharacteristicNameUserSupplied:
    return self.__characteristicNameUserSupplied
  @characteristicNameUserSupplied.setter
  def characteristicNameUserSupplied(self, val:CharacteristicNameUserSupplied) -> None:
    self.__characteristicNameUserSupplied = None if val is None else CharacteristicNameUserSupplied(val)

  @property
  def methodSpeciationName(self) -> MethodSpeciationName:
    return self.__methodSpeciationName
  @methodSpeciationName.setter
  def methodSpeciationName(self, val:MethodSpeciationName) -> None:
    self.__methodSpeciationName = None if val is None else MethodSpeciationName(val)

  @property
  def resultSampleFractionText(self) -> ResultSampleFractionText:
    return self.__resultSampleFractionText
  @resultSampleFractionText.setter
  def resultSampleFractionText(self, val:ResultSampleFractionText) -> None:
    self.__resultSampleFractionText = None if val is None else ResultSampleFractionText(val)

  @property
  def resultMeasure(self) -> Measure:
    """The reportable measure of the result for the chemical, microbiological or other characteristic being analyzed."""
    return self.__resultMeasure
  @resultMeasure.setter
  def resultMeasure(self, val:Measure) -> None:
    """The reportable measure of the result for the chemical, microbiological or other characteristic being analyzed."""
    self.__resultMeasure = None if val is None else Measure(val)

  @property
  def targetCount(self) -> TargetCount:
    return self.__targetCount
  @targetCount.setter
  def targetCount(self, val:TargetCount) -> None:
    self.__targetCount = None if val is None else TargetCount(val)

  @property
  def proportionSampleProcessedNumeric(self) -> ProportionSampleProcessedNumeric:
    return self.__proportionSampleProcessedNumeric
  @proportionSampleProcessedNumeric.setter
  def proportionSampleProcessedNumeric(self, val:ProportionSampleProcessedNumeric) -> None:
    self.__proportionSampleProcessedNumeric = None if val is None else ProportionSampleProcessedNumeric(val)

  @property
  def resultStatusIdentifier(self) -> ResultStatusIdentifier:
    return self.__resultStatusIdentifier
  @resultStatusIdentifier.setter
  def resultStatusIdentifier(self, val:ResultStatusIdentifier) -> None:
    self.__resultStatusIdentifier = None if val is None else ResultStatusIdentifier(val)

  @property
  def statisticalBaseCode(self) -> StatisticalBaseCode:
    return self.__statisticalBaseCode
  @statisticalBaseCode.setter
  def statisticalBaseCode(self, val:StatisticalBaseCode) -> None:
    self.__statisticalBaseCode = None if val is None else StatisticalBaseCode(val)

  @property
  def statisticalNValueNumeric(self) -> StatisticalNValueNumeric:
    return self.__statisticalNValueNumeric
  @statisticalNValueNumeric.setter
  def statisticalNValueNumeric(self, val:StatisticalNValueNumeric) -> None:
    self.__statisticalNValueNumeric = None if val is None else StatisticalNValueNumeric(val)

  @property
  def resultValueTypeName(self) -> ResultValueTypeName:
    return self.__resultValueTypeName
  @resultValueTypeName.setter
  def resultValueTypeName(self, val:ResultValueTypeName) -> None:
    self.__resultValueTypeName = None if val is None else ResultValueTypeName(val)

  @property
  def resultWeightBasisText(self) -> ResultWeightBasisText:
    return self.__resultWeightBasisText
  @resultWeightBasisText.setter
  def resultWeightBasisText(self, val:ResultWeightBasisText) -> None:
    self.__resultWeightBasisText = None if val is None else ResultWeightBasisText(val)

  @property
  def resultTimeBasisText(self) -> ResultTimeBasisText:
    return self.__resultTimeBasisText
  @resultTimeBasisText.setter
  def resultTimeBasisText(self, val:ResultTimeBasisText) -> None:
    self.__resultTimeBasisText = None if val is None else ResultTimeBasisText(val)

  @property
  def resultTemperatureBasisText(self) -> ResultTemperatureBasisText:
    return self.__resultTemperatureBasisText
  @resultTemperatureBasisText.setter
  def resultTemperatureBasisText(self, val:ResultTemperatureBasisText) -> None:
    self.__resultTemperatureBasisText = None if val is None else ResultTemperatureBasisText(val)

  @property
  def resultParticleSizeBasisText(self) -> ResultParticleSizeBasisText:
    return self.__resultParticleSizeBasisText
  @resultParticleSizeBasisText.setter
  def resultParticleSizeBasisText(self, val:ResultParticleSizeBasisText) -> None:
    self.__resultParticleSizeBasisText = None if val is None else ResultParticleSizeBasisText(val)

  @property
  def dataQuality(self) -> DataQuality:
    return self.__dataQuality
  @dataQuality.setter
  def dataQuality(self, val:DataQuality) -> None:
    self.__dataQuality = None if val is None else DataQuality(val)

  @property
  def resultCommentText(self) -> CommentText:
    """Free text with general comments concerning the result."""
    return self.__resultCommentText
  @resultCommentText.setter
  def resultCommentText(self, val:CommentText) -> None:
    """Free text with general comments concerning the result."""
    self.__resultCommentText = None if val is None else CommentText(val)

  @property
  def resultDepthHeightMeasure(self) -> MeasureCompact:
    """A measurement of the vertical location (measured from a reference point) at which a result is obtained."""
    return self.__resultDepthHeightMeasure
  @resultDepthHeightMeasure.setter
  def resultDepthHeightMeasure(self, val:MeasureCompact) -> None:
    """A measurement of the vertical location (measured from a reference point) at which a result is obtained."""
    self.__resultDepthHeightMeasure = None if val is None else MeasureCompact(val)

  @property
  def resultDepthAltitudeReferencePointText(self) -> DepthAltitudeReferencePointText:
    """The reference used to indicate the datum or reference used to establish the depth/altitude of a result."""
    return self.__resultDepthAltitudeReferencePointText
  @resultDepthAltitudeReferencePointText.setter
  def resultDepthAltitudeReferencePointText(self, val:DepthAltitudeReferencePointText) -> None:
    """The reference used to indicate the datum or reference used to establish the depth/altitude of a result."""
    self.__resultDepthAltitudeReferencePointText = None if val is None else DepthAltitudeReferencePointText(val)

  @property
  def resultSamplingPointName(self) -> ResultSamplingPointName:
    return self.__resultSamplingPointName
  @resultSamplingPointName.setter
  def resultSamplingPointName(self, val:ResultSamplingPointName) -> None:
    self.__resultSamplingPointName = None if val is None else ResultSamplingPointName(val)

  @property
  def resultSamplingPointType(self) -> ResultSamplingPointType:
    return self.__resultSamplingPointType
  @resultSamplingPointType.setter
  def resultSamplingPointType(self, val:ResultSamplingPointType) -> None:
    self.__resultSamplingPointType = None if val is None else ResultSamplingPointType(val)

  @property
  def resultSamplingPointPlaceInSeries(self) -> ResultSamplingPointPlaceInSeries:
    return self.__resultSamplingPointPlaceInSeries
  @resultSamplingPointPlaceInSeries.setter
  def resultSamplingPointPlaceInSeries(self, val:ResultSamplingPointPlaceInSeries) -> None:
    self.__resultSamplingPointPlaceInSeries = None if val is None else ResultSamplingPointPlaceInSeries(val)

  @property
  def resultSamplingPointCommentText(self) -> ResultSamplingPointCommentText:
    return self.__resultSamplingPointCommentText
  @resultSamplingPointCommentText.setter
  def resultSamplingPointCommentText(self, val:ResultSamplingPointCommentText) -> None:
    self.__resultSamplingPointCommentText = None if val is None else ResultSamplingPointCommentText(val)

  @property
  def recordIdentifierUserSupplied(self) -> RecordIdentifierUserSupplied:
    return self.__recordIdentifierUserSupplied
  @recordIdentifierUserSupplied.setter
  def recordIdentifierUserSupplied(self, val:RecordIdentifierUserSupplied) -> None:
    self.__recordIdentifierUserSupplied = None if val is None else RecordIdentifierUserSupplied(val)

  def generateXML(self, name:str = 'ResultDescription') -> str:
    doc = Doc()
    asis = doc.asis
    line = doc.line
    tag = doc.tag

    with tag(name):
      if self.__dataLoggerLineName is not None:
        line('DataLoggerLineName', self.__dataLoggerLineName)
      if self.__resultDetectionConditionText is not None:
        line('ResultDetectionConditionText', self.__resultDetectionConditionText)
      if self.__characteristicName is not None:
        line('CharacteristicName', self.__characteristicName)
      if self.__characteristicNameUserSupplied is not None:
        line('CharacteristicNameUserSupplied', self.__characteristicNameUserSupplied)
      if self.__methodSpeciationName is not None:
        line('MethodSpeciationName', self.__methodSpeciationName)
      if self.__resultSampleFractionText is not None:
        line('ResultSampleFractionText', self.__resultSampleFractionText)
      if self.__resultMeasure is not None:
        asis(self.__resultMeasure.generateXML('ResultMeasure'))
      if self.__targetCount is not None:
        line('TargetCount', self.__targetCount)
      if self.__proportionSampleProcessedNumeric is not None:
        line('ProportionSampleProcessedNumeric', self.__proportionSampleProcessedNumeric)
      if self.__resultStatusIdentifier is not None:
        line('ResultStatusIdentifier', self.__resultStatusIdentifier)
      if self.__statisticalBaseCode is not None:
        line('StatisticalBaseCode', self.__statisticalBaseCode)
      if self.__statisticalNValueNumeric is not None:
        line('StatisticalNValueNumeric', self.__statisticalNValueNumeric)
      if self.__resultValueTypeName is not None:
        line('ResultValueTypeName', self.__resultValueTypeName)
      if self.__resultWeightBasisText is not None:
        line('ResultWeightBasisText', self.__resultWeightBasisText)
      if self.__resultTimeBasisText is not None:
        line('ResultTimeBasisText', self.__resultTimeBasisText)
      if self.__resultTemperatureBasisText is not None:
        line('ResultTemperatureBasisText', self.__resultTemperatureBasisText)
      if self.__resultParticleSizeBasisText is not None:
        line('ResultParticleSizeBasisText', self.__resultParticleSizeBasisText)
      if self.__dataQuality is not None:
        asis(self.__dataQuality.generateXML('DataQuality'))
      if self.__resultCommentText is not None:
        line('ResultCommentText', self.__resultCommentText)
      if self.__resultDepthHeightMeasure is not None:
        asis(self.__resultDepthHeightMeasure.generateXML('ResultDepthHeightMeasure'))
      if self.__resultDepthAltitudeReferencePointText is not None:
        line('ResultDepthAltitudeReferencePointText', self.__resultDepthAltitudeReferencePointText)
      if self.__resultSamplingPointName is not None:
        line('ResultSamplingPointName', self.__resultSamplingPointName)
      if self.__resultSamplingPointType is not None:
        line('ResultSamplingPointType', self.__resultSamplingPointType)
      if self.__resultSamplingPointPlaceInSeries is not None:
        line('ResultSamplingPointPlaceInSeries', self.__resultSamplingPointPlaceInSeries)
      if self.__resultSamplingPointCommentText is not None:
        line('ResultSamplingPointCommentText', self.__resultSamplingPointCommentText)
      if self.__recordIdentifierUserSupplied is not None:
        line('RecordIdentifierUserSupplied', self.__recordIdentifierUserSupplied)

    return doc.getvalue()
